"""
Utilities for creating and mixing sine wave signals.

Provides functions to generate and combine sine waves with different frequencies,
phases, and amplitudes to create complex waveforms.
"""

from numpy import sin, arange, pi, ones, ndarray, random, all
from loopyng.gen.util import DFLT_N_SAMPLES, DFLT_SR, DFLT_FREQ


def mk_sine_wf(freq=DFLT_FREQ, n_samples=DFLT_N_SAMPLES, sr=DFLT_SR, phase=0, gain=1):
    """Make a sine waveform

    :param freq: Frequency (in Hz)
    :param n_samples: The number of samples of waveform you want
    :param sr: Sample rate
    :param phase: Phase (in radians)
    :param gain: (A number to multiply the base sine wave by)
    :return: Waveform. A numpy array of samples of the specified sine wave

    >>> n_samples = random.randint(2,5)
    >>> wf = mk_sine_wf(n_samples=n_samples)
    >>> assert len(wf) == n_samples
    >>> wf = mk_sine_wf(n_samples=3)
    >>> import numpy as np
    >>> assert np.allclose(wf, [0.        , 0.06264832, 0.12505052], rtol=1e-3)
    """
    return gain * sin(phase + arange(n_samples) * 2 * pi * freq / sr)


def freq_based_stationary_wf(
    freqs=(200, 400, 600, 800),
    weights=None,
    n_samples: int = DFLT_N_SAMPLES,
    sr: int = DFLT_SR,
) -> ndarray:
    """
    Makes a stationary waveform by mixing a number of freqs together,
    possibly with different weights.

    :param freqs: List(-like) of frequencies (in Hz)
    :param weights: The weights these frequencies should have (all weights will be normalized
    :param n_samples: The number of samples of waveform you want
    :param sr: Sample rate
    :return: Waveform. A numpy array of samples of the specified sine wave

    >>> n_samples = random.randint(2,5)
    >>> wf = freq_based_stationary_wf(n_samples=n_samples)
    >>> assert len(wf) == n_samples
    >>> wf = freq_based_stationary_wf(n_samples = 3, weights = [1,2,3,4])
    >>> import numpy as np
    >>> assert np.allclose(wf, [0.        , 0.08534908, 0.16988139], rtol=1e-3)
    """
    if weights is None:
        weights = ones(len(freqs))
    assert len(freqs) == len(weights)
    _mk_sine_wf = partial(mk_sine_wf, n_samples=n_samples, sr=sr)
    wf = sum(_mk_sine_wf(freq) * weights[i] for i, freq in enumerate(freqs))
    return wf / sum(weights)


#################################################################################
# Soft marking everything below for deprecation
import random
from typing import Callable, Union, Tuple, Optional, Iterable
from functools import partial
from dataclasses import dataclass
from numbers import Number
from itertools import islice


@dataclass
class MinMaxRand:
    """Like a partial, but meant for bounded rand functions.
    Could have just done:

    ```
    r = partial(random.uniform, min, max)
    r()
    ```

    But we need to be able to have access to r.min and r.max
    """

    min: Number = 0
    max: Number = 1
    rand_func: Callable[[Number, Number], Number] = random.uniform

    def __call__(self):
        return self.rand_func(self.min, self.max)


@dataclass
class MinMaxRandDict:
    """
    >>> from loopyng.gen.sine_mix import MinMaxRandDict, MinMaxRand
    >>>
    >>> rand_gen = MinMaxRandDict((
    ...     ('rpm', MinMaxRand(100, 1000)),
    ...     ('temperature', MinMaxRand(10, 25)),
    ...     ('pressure', MinMaxRand(50, 500)),
    ... ))
    >>> t = list(rand_gen.read(4))
    >>> assert len(t) == 4
    >>> assert list(t[0]) == ['rpm', 'temperature', 'pressure']
    >>> t  # doctest: +SKIP
    [{'rpm': 577.8533852127333,
      'temperature': 18.306495707512724,
      'pressure': 357.72481026748113},
     {'rpm': 457.0605450014562,
      'temperature': 23.94969543436332,
      'pressure': 483.1119334545017},
     {'rpm': 489.8445225473431,
      'temperature': 17.176275589569695,
      'pressure': 415.8337855923849},
     {'rpm': 120.84000962564362,
      'temperature': 21.153044572476936,
      'pressure': 449.5124385844328}]
    """

    iid_seed_gen: Union[dict, Tuple[Tuple[str, MinMaxRand]]] = (
        ("rpm", MinMaxRand(100, 1000)),
        ("temperature", MinMaxRand(10, 25)),
        ("pressure", MinMaxRand(50, 500)),
    )

    def __post_init__(self):
        self.iid_seed_gen = dict(self.iid_seed_gen)

    def __call__(self, n=1):
        if n == 1:
            return {k: v() for k, v in self.iid_seed_gen.items()}
        else:
            return self.read(n)

    def __iter__(self):
        while True:
            yield self()

    def read(self, n=1):
        return islice(self, n)


def dflt_wf_params_to_wf(
    weights,
    freqs=None,
    n_samples: int = DFLT_N_SAMPLES,
    sr: int = DFLT_SR,
):
    """A default for wf_params_to_wf

    :param weights: The weights these frequencies should have (all weights will be normalized
    :param freqs: List(-like) of frequencies (in Hz)
    :param n_samples: The number of samples of waveform you want
    :param sr: Sample rate
    :return: Waveform. A numpy array of samples of the specified sine wave

    >>> n_samples = random.randint(2,5)
    >>> wf = dflt_wf_params_to_wf(weights = [1,2,3,4], n_samples=100)
    >>> len(wf)
    100
    >>> wf  # doctest: +SKIP
    array([ 0.        ,  0.08534908,  0.16988139,  ..., -0.22847099, -0.22552675])

    """
    if freqs is None:
        freqs = tuple(range(200, (len(weights) + 1) * 200, 200))
    return freq_based_stationary_wf(freqs, weights, n_samples, sr)


def dflt_sprout_to_wf_params(annot):
    # TODO: Need to include the possibility of normalization
    params = tuple(annot.values())
    return params


def asis(x):
    return x


# def numerical_annotations_and_waveform_chunks(


# TODO: Replicate in slink
@dataclass
class NumAnnotsAndWaveformChunks:
    seeds: Union[dict, Tuple[Tuple[str, MinMaxRand]]] = (
        ("rpm", MinMaxRand(100, 1000)),
        ("temperature", MinMaxRand(10, 25)),
        ("pressure", MinMaxRand(50, 500)),
    )
    seed_to_sprout_gen: Callable = MinMaxRandDict
    sprout_to_annot: Callable = asis
    sprout_to_wf_params: Callable = dflt_sprout_to_wf_params
    wf_params_to_wf: Callable = dflt_wf_params_to_wf

    def __post_init__(self):
        self.sprout_gen = self.seed_to_sprout_gen(self.seeds)

    def __call__(self, *args, **kwargs):
        sprout = self.sprout_gen()
        annot = self.sprout_to_annot(sprout)
        params = self.sprout_to_wf_params(annot)
        wf = self.wf_params_to_wf(params)
        return annot, wf

    def __iter__(self):
        while True:
            yield self()

    def read(self, n=1):
        return islice(self, n)


##########



"""
Utils to view, hear, and manipulate audio.
"""

from numpy import (
    array,
    mean,
    shape,
    linspace,
    max,
    log10,
    ceil,
    int16,
    hstack,
    zeros,
    argmin,
    ndim,
)
from numpy.random import randint
import soundfile as sf
import os
import matplotlib.pylab as plt
from IPython.display import Audio

from loopyng.utils.plotting import plot_wf

try:
    import librosa
    import librosa.display

    WITH_LIBROSA = True
except ImportError:
    WITH_LIBROSA = False
    import warnings

    warnings.warn(
        "Couldn't import librosa. You can install is, but know that you "
        "don't NEED it unless you want to compute melspectrograms."
    )

    class librosa:
        def __getattr__(self, a):
            raise ModuleNotFoundError(
                f"You don't actually have {self.__class__.__name__}"
            )


default_sr = 44100
default_wf_type = int16


def subtype_of_wf(wf):
    if len(wf) > 0:
        if isinstance(wf[0], (int, int16)):
            return "PCM_16"
        else:
            return "FLOAT"
    else:
        return None


def stereo_to_mono_by_taking_first_channel(wf):
    if ndim(wf) == 1:
        return wf
    else:
        likely_channel_dim = argmin(wf.shape)
        if likely_channel_dim == 0:
            return wf[0, :]
        else:
            return wf[:, 0]


def ensure_channels_in_columns(wf):
    likely_channel_dim = argmin(wf.shape)
    if likely_channel_dim == 0:
        return wf.T
    else:
        return wf


def ensure_mono(wf):
    if ndim(wf) == 1:
        return wf
    else:
        return stereo_to_mono_by_taking_first_channel(wf)


def is_wav_file(filepath):
    return os.path.splitext(filepath)[1] == ".wav"


def plot_melspectrogram(spect_mat, sr=default_sr, hop_length=512, name=None):
    # Make a new figure
    plt.figure(figsize=(12, 4))
    # Display the spectrogram on a mel scale
    # sample rate and hop length parameters are used to render the time axis
    librosa.display.specshow(
        spect_mat, sr=sr, hop_length=hop_length, x_axis="time", y_axis="mel"
    )
    # Put a descriptive title on the plot
    if name is not None:
        plt.title('mel power spectrogram of "{}"'.format(name))
    else:
        plt.title("mel power spectrogram")
    # draw a color bar
    plt.colorbar(format="%+02.0f dB")
    # Make the figure layout compact
    plt.tight_layout()


def wf_and_sr(*args, **kwargs):
    if len(args) > 0:
        args_0 = args[0]
        if isinstance(args_0, str):
            kwargs["filepath"] = args_0
        elif isinstance(args_0, tuple):
            kwargs["wf"], kwargs["sr"] = args_0
    kwargs_keys = list(kwargs.keys())
    if "wf" in kwargs_keys:
        return kwargs["wf"], kwargs["sr"]


class Sound(object):
    def __init__(self, wf=None, sr=default_sr, wf_type=default_wf_type):
        if wf is None:
            wf = array([], dtype=wf_type)
        self.wf = wf
        self.sr = sr

    def copy(self):
        return self.__class__(wf=self.wf.copy(), sr=self.sr)

    def __len__(self):
        return len(self.wf)

    def convert_interval_to_samples_unit(self, interval):
        if isinstance(interval, tuple) and len(interval) <= 3:
            interval = slice(*interval)
        if (
            isinstance(interval.start, float)
            or isinstance(interval.stop, float)
            or isinstance(interval.step, float)
        ):
            if interval.start is not None:
                _start = int(interval.start * self.sr)
            else:
                _start = None
            if interval.stop is not None:
                _stop = int(interval.stop * self.sr)
            else:
                _stop = None
            if interval.step is not None:
                _step = int(interval.step * self.sr)
            else:
                _step = None
            interval = slice(_start, _stop, _step)

        return interval

    def __getitem__(self, item):
        item = self.convert_interval_to_samples_unit(item)
        return self.__class__(self.wf.__getitem__(item), self.sr)

    @property
    def duration_s(self):
        """
        Deprecated: Use chk_size_ms instead. Just included here for back-compatibility.
        :return: duration, in seconds, of the sound
        """
        return len(self.wf) / self.sr

    ####################################################################################################################
    # CREATION

    @classmethod
    def from_file(cls, filepath, wf_type=default_wf_type):
        """
        Construct sound object from a wav audio file
        :param filepath: filepath of the sound file
        :param wf_type: type of the wf numbers
        :return: a Sound object
        """
        # kwargs = dict({'always_2d': False, 'ensure_mono': True}, **kwargs)

        wf, sr = sf.read(filepath, dtype=wf_type)
        wf = ensure_mono(wf)
        return cls(wf=wf, sr=sr)

    @classmethod
    def from_(cls, sound):
        if isinstance(sound, tuple) and len(sound) == 2:  # then it's a (wf, sr) tuple
            return cls(sound[0], sound[1])
        elif isinstance(sound, str) and os.path.isfile(sound):
            return cls.from_file(sound)
        elif isinstance(sound, dict):
            if "wf" in sound and "sr" in sound:
                return cls(sound["wf"], sound["sr"])
            else:
                return cls.from_sref(sound)
        elif hasattr(sound, "wf") and hasattr(sound, "sr"):
            return cls(sound.wf, sound.sr)
        else:
            raise TypeError("Couldn't figure out how that format represents sound")

    @classmethod
    def silence(cls, seconds=0.0, sr=default_sr, wf_type=default_wf_type):
        return cls(sr=sr, wf=zeros(int(round(seconds * sr)), wf_type))

    def save_to_wav(self, filepath=None, samplerate=None, **kwargs):
        subtype = kwargs.get("subtype", subtype_of_wf(self.wf))
        samplerate = samplerate or self.sr
        if isinstance(filepath, int):
            rand_range = filepath
            template = "sound_save_{:0" + str(int(ceil(log10(rand_range)))) + ".0}.wav"
            filepath = template.format(randint(0, rand_range))
        else:
            filepath = filepath or "sound_save.wav"
        sf.write(filepath, self.wf, samplerate=samplerate, subtype=subtype, **kwargs)

    ####################################################################################################################
    # TRANSFORMATIONS

    def ensure_mono(self):
        self.wf = ensure_mono(self.wf)

    def crop_with_idx(self, first_idx, last_idx):
        cropped_sound = self.copy()
        cropped_sound.wf = cropped_sound.wf[first_idx : (last_idx + 1)]
        return cropped_sound

    def crop_with_seconds(self, first_second, last_second):
        return self.crop_with_idx(
            int(round(first_second * self.sr)),
            int(round(last_second * self.sr)),
        )

    def melspectr_matrix(self, **mel_kwargs):
        mel_kwargs = dict(
            {"n_fft": 2048, "hop_length": 512, "n_mels": 128}, **mel_kwargs
        )
        S = librosa.feature.melspectrogram(
            array(self.wf).astype(float), sr=self.sr, **mel_kwargs
        )
        # Convert to log scale (dB). We'll use the peak power as reference.
        return librosa.amplitude_to_db(S, ref=max)

    def __add__(self, append_sound):
        assert (
            self.sr == append_sound.sr
        ), "Sounds need to have the same sample rate to be appended"
        return Sound(sr=self.sr, wf=hstack(self.wf, append_sound.wf))

    ####################################################################################################################
    # DISPLAY FUNCTIONS

    def hear(self, autoplay=False, **kwargs):
        wf = array(ensure_mono(self.wf)).astype(float)
        wf[
            randint(len(wf))
        ] *= 1.001  # hack to avoid having exactly the same sound twice (creates an Audio bug)
        return Audio(data=wf, rate=self.sr, autoplay=autoplay, **kwargs)

    def plot_wf(*args, **kwargs):
        wf, sr = wf_and_sr(*args, **kwargs)
        return plot_wf(wf, sr)

    def display(self, autoplay=False, **kwargs):
        """
        :param sound_plot: 'mel' (default) to plot melspectrogram, 'wf' to plot wave form, and None to plot nothing at all
        :param kwargs:
        :return:
        """
        if WITH_LIBROSA:
            self.melspectrogram(plot_it=True, **kwargs)
        else:
            self.plot_wf()

        return self.hear(autoplay=autoplay)

    if WITH_LIBROSA:

        def melspectrogram(self, plot_it=False, **mel_kwargs):
            mel_kwargs = dict(
                {"n_fft": 2048, "hop_length": 512, "n_mels": 128}, **mel_kwargs
            )
            log_S = self.melspectr_matrix(**mel_kwargs)
            if plot_it:
                plot_melspectrogram(
                    log_S, sr=self.sr, hop_length=mel_kwargs["hop_length"]
                )
            return log_S

    else:

        def melspectrogram(self, plot_it=False, **mel_kwargs):
            if plot_it:
                plt.figure(figsize=(16, 5))
                plt.plot(self.wf)



"""
Date tick formatting utilities for time-based plots.

Provides functions to convert numeric timestamps to formatted time strings for plot axes.
"""

import numpy as np
from datetime import datetime as dt

unit_str_to_unit_in_seconds = {
    "day": 3600 * 24,
    "hour": 3600,
    "h": 3600,
    "minute": 60,
    "mn": 60,
    "second": 1,
    "s": 1,
    "millisecond": 0.001,
    "ms": 0.001,
    "microsecond": 1e-6,
    "us": 1e-6,
}

unit_in_seconds = np.array(
    [
        60 * 60 * 24 * 365,  # year
        60 * 60 * 24 * 30,  # month
        60 * 60 * 24 * 7,  # week
        60 * 60 * 24,  # day
        60 * 60,  # hour
        60,  # minute
        1,  # second
        1e-3,  # millisecond
        1e-6,  # microsecond
        1e-9,  # nanosecond
    ]
)

strftime_format_for_unit = {
    60 * 60 * 24 * 30: "%y-%m-%d",  # month
    60 * 60 * 24 * 7: "%b %d",  # week
    60 * 60 * 24: "%b %d",  # day
    60 * 60: "%d-%H:%M",  # hour
    60: "%H:%M:%S",  # minute
    1: "%M:%S.%f",  # second
    1e-3: "%S''%f",  # millisecond
    1e-6: "%S''%f",  # microsecond
}


def utc_datetime_from_val_and_unit(val, unit):
    if isinstance(unit, str):
        unit = unit_str_to_unit_in_seconds[unit]
    return dt.utcfromtimestamp(val * unit)


def largest_unit_that_changes_at_every_tick(ticks, ticks_unit):
    """
    Returns the largest time unit for which each time tick changes.
    :param ticks: The list of ticks
    :param ticks_unit: The unit of the elements of ticks, expressed in seconds. For example, if the list
        contains hours, unit=3600, if minutes, unit=60, if seconds unit=1,
        if milliseconds unit=0.001.
            Note: You can also use a string to express the unit, as long as it's recognized by the
            unit_str_to_unit_in_seconds dict. Keys recognized:
            ['day', 'hour', 'h', 'minute', 'mn', 'second', 's', 'millisecond', 'ms', 'microsecond', 'us']
    :return:
    """
    ticks = np.array(ticks)
    if isinstance(ticks_unit, str):
        ticks_unit = unit_str_to_unit_in_seconds[ticks_unit]
    min_tick_diff = min(np.diff(ticks))
    min_tick_diff *= ticks_unit  # convert to seconds

    for u in unit_in_seconds:
        if u < min_tick_diff:
            return u


def strftime_format_for_ticks(ticks, ticks_unit):
    unit = largest_unit_that_changes_at_every_tick(ticks, ticks_unit)
    return strftime_format_for_unit[unit]


def strftime_with_precision(tick, format, sub_secs_precision=2):
    """
    Returns a formatted string for a datetime (tick).
    :param tick: The datetime for this tick
    :param format: The formatting string
    :param sub_secs_precision: Number of digits to used for sub-seconds.
        If None, will choose it "smartly/dynamically"
    :return: Formatted string
    """
    t = tick.strftime(format)
    is_us = "%f" in format
    if is_us:
        if sub_secs_precision is None:
            while t[-1] == "0":
                t = t[:-1]
            while not t[-1].isdigit():
                t = t[:-1]
            return t
        else:
            if sub_secs_precision < 0:
                sub_secs_precision = 0
            elif sub_secs_precision > 6:
                sub_secs_precision = 6

            DFLT_PRECISION = 6
            digits_to_skip = DFLT_PRECISION - sub_secs_precision
            if digits_to_skip == 0:
                return t
            else:
                t = t[:-digits_to_skip]
                while not t[-1].isdigit():
                    t = t[:-1]
                return t
    else:
        return t


def str_ticks(ticks, ticks_unit, sub_secs_precision=2):
    t_format = strftime_format_for_ticks(ticks, ticks_unit)
    return [
        strftime_with_precision(
            utc_datetime_from_val_and_unit(x, ticks_unit),
            t_format,
            sub_secs_precision,
        )
        for x in ticks
    ]


def unit_aligned_ticks(ticks, ticks_unit):
    pass



import random
from typing import Callable, Union
from functools import partial

from more_itertools import distribute, minmax
import numpy as np

from i2 import Pipe

from loopyng.gen.sine_mix import dflt_wf_params_to_wf
from loopyng.util import simple_chunker


DFLT_SR = 44100
DFLT_FACTOR = 100
DFLT_TAG_MODEL = {
    "normal": lambda last_item: {
        "bt": (last_item["tt"] if last_item else 0)
        + abs(random.gauss(mu=0, sigma=0.5)),
        "duration": 0.1 + abs(random.gauss(mu=0, sigma=1)),
    },
    "abnormal": lambda last_item: {
        "bt": (last_item["tt"] if last_item else 0)
        + abs(random.gauss(mu=0.5, sigma=1)),
        "duration": 0.05 + abs(random.gauss(mu=0, sigma=0.25)),
    },
}

rpm = Pipe(partial(random.uniform, 400, 800), int)


def mean_crossing_count(arr):
    mean_ = np.mean(arr)
    return int(sum(np.diff(np.array(arr) > mean_)))


def frequs_from_cats(cats, num_freq_per_cat=2):
    groups = distribute(
        num_freq_per_cat,
        tuple(range(200, num_freq_per_cat * (len(cats) + 1) * 200, 200)),
    )

    return {cat: list(group) for cat, group in zip(cats, groups)}


def tag_model_to_params(tag_model):
    keys = list(tag_model.keys())
    return frequs_from_cats(keys)


def interval_to_duration(interval, sr=DFLT_SR):
    return (interval["tt"] - interval["bt"]) * sr


def identity(x):
    return x


def tagged_intervals_gen(tag_model=None, items_per_tag=None, start_bt_s=0):
    """
    Generates a sequence of tagged intervals in the form of dicts
    of the form {"tag": str, "bt": numerical, "tt": numerical} where bt is
    the lower bound of the interval and tt is the upper bound of the interval.

    :param tag_model: A dict whose keys are the tags and values are the generative
                      model for that tag.
    :param n_items: Number of items to be generated for each tag. If None, one
                    item for each tag will be generated.
    :param start_bt_s: The start time in seconds.
    :return: A generator yielding dicts representing tagged intervals.
    """
    if tag_model is None:
        tag_model = DFLT_TAG_MODEL

    if items_per_tag is None:
        items_per_tag = {tag: 1 for tag in tag_model}
    elif isinstance(items_per_tag, int):
        items_per_tag = {tag: items_per_tag for tag in tag_model}

    last_items = {tag: None for tag in tag_model}

    for tag, generator in tag_model.items():
        for _ in range(items_per_tag[tag]):
            generated = generator(last_items[tag])
            bt = max(
                generated["bt"],
                (last_items[tag]["tt"] if last_items[tag] else start_bt_s),
            )
            tt = bt + generated["duration"]
            last_items[tag] = {"bt": bt, "tt": tt}
            yield {"tag": tag, "bt": bt, "tt": tt}


def intervals_to_wf(intervals, sr=DFLT_SR, rescaler=identity):
    _, max_tt = minmax(intervals, key=lambda x: x["tt"])
    end_wf = int(max_tt["tt"] * sr)
    wf = np.zeros(end_wf + 1)
    for interval in intervals:
        wf_params = tag_model_to_params(tag_model=DFLT_TAG_MODEL)[interval["tag"]]
        n_samples = int(interval_to_duration(interval))
        bt = int(interval["bt"] * sr)
        tt = bt + n_samples
        wf[bt:tt] = dflt_wf_params_to_wf(wf_params, n_samples=n_samples, sr=sr)
    return wf


def intervals_to_plc(intervals, sr=DFLT_SR, factor=DFLT_FACTOR, rescaler=identity):
    _, max_tt = minmax(intervals, key=lambda x: x["tt"])
    end_wf = int(max_tt["tt"] * sr)
    wf = np.zeros(end_wf + 1)
    for interval in intervals:
        delta = 0
        n_samples = int(interval_to_duration(interval))
        bt = int(interval["bt"] * sr)
        if interval["tag"] == "normal":
            delta = 400
        vals = ([delta + rpm()] * factor for _ in range(n_samples // factor))
        flattened = [item for sublist in vals for item in sublist]
        wf[bt : bt + len(flattened)] = flattened

    # TODO: This is just a hack to make the PLC have a different sr than wf
    #  --> Remove when the proper PLC generator is implemented
    chks = simple_chunker(wf, chk_size=factor)
    wf = list(map(mean_crossing_count, chks))

    return wf


def mk_channel_data(channel, data, ts, sr):
    return {"channel": channel, "data": list(data), "ts": ts, "sr": sr}


def rescale_intervals(intervals, rescaler):
    return [
        {
            "tag": interval["tag"],
            "bt": rescaler(interval["bt"]),
            "tt": rescaler(interval["tt"]),
        }
        for interval in intervals
    ]


def intervals_to_json(intervals, sr=DFLT_SR, factor=DFLT_FACTOR, rescaler=identity):
    import json

    result = dict()
    result["data"] = list()
    wf = intervals_to_wf(intervals, rescaler=rescaler)
    plc = intervals_to_plc(intervals, rescaler=rescaler)

    first_ts = rescaler(intervals[0]["bt"])
    result["data"].append(mk_channel_data("wf", wf, first_ts, sr))
    result["data"].append(mk_channel_data("plc", plc, first_ts, sr // factor))
    result["data"].append(
        {
            "channel": "annot",
            "data": rescale_intervals(intervals, rescaler),
        }
    )
    volume_list, mixed_list = mk_vol_mixed_data(intervals, wf, rescaler)
    result["data"].append({"channel": "volume", "data": volume_list})
    result["data"].append({"channel": "mixed", "data": mixed_list})

    return json.dumps(result)


def mk_vol_mixed_data(intervals, wf, rescaler=identity):
    volume_list = list()
    mixed_list = list()
    for interval in intervals:
        wf_chunk = read_chk(wf, interval)
        volume = np.std(wf_chunk)
        mean = np.mean(wf_chunk)
        volume_list.append({"value": volume, "ts": rescaler(interval["bt"])})
        mixed_list.append(
            {"values": {"mean": mean, "std": volume}, "ts": rescaler(interval["bt"])}
        )

    return volume_list, mixed_list


def read_chk(wf, interval, sr=DFLT_SR):
    wf_chunk = wf[int(interval["bt"] * sr) : int(interval["tt"] * sr)]
    return wf_chunk


def thin_out(array, sample_size: Union[int, float] = 0.5):
    """Thin out an array by randomly selecting a sample of the array."""
    array_len = len(array)
    if sample_size < 1:
        sample_size = array_len * sample_size
    sample_size = int(sample_size)
    indices = sorted(np.random.choice(array_len, sample_size, replace=False))
    return np.array(array)[indices]


sample_offset_with_dflt_sr = lambda x: int(x * DFLT_SR)

_rescalers = dict(
    identity=identity,
    sample_offset_with_dflt_sr=sample_offset_with_dflt_sr,
)


# TODO: Emulating the existing code here, but should separate json concern from the rest


def five_channel_json_data(
    n_generators: Union[float, int] = 1.0,
    *,
    tag_model=DFLT_TAG_MODEL,
    items_per_tag: int = 3,
    start_bt_s: Union[float, int] = 0,
    rescaler: Union[Callable, str] = sample_offset_with_dflt_sr,
):
    """Generate a json string with 5 channels of data."""
    print(f"--------------- {rescaler}")

    if isinstance(rescaler, str):
        if rescaler not in _rescalers:
            raise ValueError(
                f"Unknown rescaler: {rescaler}. Valid values: {list(_rescalers)}"
            )
        rescaler = _rescalers[rescaler]

    assert callable(rescaler)

    intervals = list(
        tagged_intervals_gen(
            tag_model=tag_model,
            items_per_tag=items_per_tag,
            start_bt_s=start_bt_s,
        )
    )

    intervals = list(thin_out(intervals, sample_size=n_generators))

    json_str = intervals_to_json(intervals, rescaler=rescaler)
    return json_str


if __name__ == "__main__":
    import argh

    argh.dispatch_command(five_channel_json_data)



"""
Generate Audio
"""

from contextlib import suppress

from loopyng.gen.util import DFLT_N_SAMPLES, DFLT_SR

from loopyng.gen.simple import inverted_tiles_wf, inverted_tiles

with suppress(ModuleNotFoundError, ImportError):
    from loopyng.gen.diagnosis_sounds import (
        WfGen,
        TimeSound,
        BinarySound,
        slow_mask,
        mk_some_buzz_wf,
        wf_with_timed_bleeps,
        mk_sounds_with_timed_bleeps,
    )

with suppress(ModuleNotFoundError, ImportError):
    from loopyng.gen.sine_mix import (
        mk_sine_wf,
        freq_based_stationary_wf,
        NumAnnotsAndWaveformChunks,
    )

with suppress(ModuleNotFoundError, ImportError):
    from loopyng.gen.voiced_time import Voicer, tell_time_continuously

with suppress(ModuleNotFoundError, ImportError):
    from loopyng.gen.signal_generators import (
        normal_dist,
        gen_words,
        categorical_gen,
        alphabet_to_bins,
        call_repeatedly,
        bernoulli,
        bernoulli_gen,
        inlier_outlier,
        signal,
        create_session,
        string_to_num,
        session_to_df,
    )



"""
Tools to create audio
"""

from loopyng.gen import *
from loopyng.utils import *
from loopyng.sound import *
from loopyng.lite_sample_sounds import (
    tag_wf_gen,
    AnnotatedWaveform,
    chk_from_pattern,
    random_samples,
    pure_tone,
    triangular_tone,
    square_tone,
    tag_to_wf_gen_func,
    mk_audio_store,
)



"""
Audio processing and playback utilities.

Provides classes and functions for working with audio data, including
playback, visualization, and basic audio transformations.
"""

from numpy import array, max, log10, ceil, int16, hstack, zeros, argmin, ndim
from numpy.random import randint
import soundfile as sf
import os
import matplotlib.pylab as plt
from IPython.display import Audio

from loopyng.utils.plotting import plot_wf
from loopyng.utils.librosa_utils import specshow, melspectrogram, amplitude_to_db


default_sr = 44100
default_wf_type = int16


def subtype_of_wf(wf):
    if len(wf) > 0:
        if isinstance(wf[0], (int, int16)):
            return "PCM_16"
        else:
            return "FLOAT"
    else:
        return None


def stereo_to_mono_by_taking_first_channel(wf):
    if ndim(wf) == 1:
        return wf
    else:
        likely_channel_dim = argmin(wf.shape)
        if likely_channel_dim == 0:
            return wf[0, :]
        else:
            return wf[:, 0]


def ensure_channels_in_columns(wf):
    likely_channel_dim = argmin(wf.shape)
    if likely_channel_dim == 0:
        return wf.T
    else:
        return wf


def ensure_mono(wf):
    if ndim(wf) == 1:
        return wf
    else:
        return stereo_to_mono_by_taking_first_channel(wf)


def is_wav_file(filepath):
    return os.path.splitext(filepath)[1] == ".wav"


def plot_melspectrogram(spect_mat, sr=default_sr, hop_length=512, name=None):
    # Make a new figure
    plt.figure(figsize=(12, 4))
    # Display the spectrogram on a mel scale
    # sample rate and hop length parameters are used to render the time axis
    specshow(spect_mat, sr=sr, hop_length=hop_length, x_axis="time", y_axis="mel")
    # Put a descriptive title on the plot
    if name is not None:
        plt.title('mel power spectrogram of "{}"'.format(name))
    else:
        plt.title("mel power spectrogram")
    # draw a color bar
    plt.colorbar(format="%+02.0f dB")
    # Make the figure layout compact
    plt.tight_layout()


def wf_and_sr(*args, **kwargs):
    """
    :param args: Either the file path to a sound or a tuple of (wf, sr)
    :param kwargs: wf = wf, sr = sr
    :return: wf, sr
    """
    if len(args) > 0:
        args_0 = args[0]
        if isinstance(args_0, str):
            kwargs["filepath"] = args_0
        elif isinstance(args_0, tuple) and len(args_0) == 2:
            kwargs["wf"], kwargs["sr"] = args_0
        elif hasattr(args_0, "wf") and hasattr(args_0, "sr"):
            kwargs["wf"], kwargs["sr"] = args_0.wf, args_0.sr
    kwargs_keys = list(kwargs.keys())
    if "wf" in kwargs_keys:
        return kwargs["wf"], kwargs["sr"]
    else:
        raise ValueError(
            f"Could not result waveform from the arguments provided: {args}, {kwargs}"
        )


class Sound(object):
    def __init__(self, wf=None, sr=default_sr, wf_type=default_wf_type):
        if wf is None:
            wf = array([], dtype=wf_type)
        self.wf = wf
        self.sr = sr

    def copy(self):
        return self.__class__(wf=self.wf.copy(), sr=self.sr)

    def __len__(self):
        return len(self.wf)

    def convert_interval_to_samples_unit(self, interval):
        if isinstance(interval, tuple) and len(interval) <= 3:
            interval = slice(*interval)
        if (
            isinstance(interval.start, float)
            or isinstance(interval.stop, float)
            or isinstance(interval.step, float)
        ):
            if interval.start is not None:
                _start = int(interval.start * self.sr)
            else:
                _start = None
            if interval.stop is not None:
                _stop = int(interval.stop * self.sr)
            else:
                _stop = None
            if interval.step is not None:
                _step = int(interval.step * self.sr)
            else:
                _step = None
            interval = slice(_start, _stop, _step)

        return interval

    def __getitem__(self, item):
        item = self.convert_interval_to_samples_unit(item)
        return self.__class__(self.wf.__getitem__(item), self.sr)

    @property
    def duration_s(self):
        """
        Deprecated: Use chk_size_ms instead. Just included here for back-compatibility.
        :return: duration, in seconds, of the sound
        """
        return len(self.wf) / self.sr

    ####################################################################################################################
    # CREATION

    @classmethod
    def from_file(cls, filepath, wf_type=default_wf_type):
        """
        Construct sound object from a wav audio file
        :param filepath: filepath of the sound file
        :param wf_type: type of the wf numbers
        :return: a Sound object
        """
        # kwargs = dict({'always_2d': False, 'ensure_mono': True}, **kwargs)

        wf, sr = sf.read(filepath, dtype=wf_type)
        wf = ensure_mono(wf)
        return cls(wf=wf, sr=sr)

    @classmethod
    def from_(cls, sound):
        """
        Construct sound object from another sound object
        """
        if isinstance(sound, tuple) and len(sound) == 2:  # then it's a (wf, sr) tuple
            return cls(sound[0], sound[1])
        elif isinstance(sound, str) and os.path.isfile(sound):
            return cls.from_file(sound)
        elif isinstance(sound, dict):
            if "wf" in sound and "sr" in sound:
                return cls(sound["wf"], sound["sr"])
            else:
                return cls.from_sref(sound)
        elif hasattr(sound, "wf") and hasattr(sound, "sr"):
            return cls(sound.wf, sound.sr)
        else:
            raise TypeError("Couldn't figure out how that format represents sound")

    @classmethod
    def silence(cls, seconds=0.0, sr=default_sr, wf_type=default_wf_type):
        """
        Construct silent sound object with length seconds

        >>> assert sum(sum(Sound.silence(seconds = 1).melspectr_matrix())) == 0.0  # doctest: +SKIP
        """
        return cls(sr=sr, wf=zeros(int(round(seconds * sr)), wf_type))

    def save_to_wav(self, filepath=None, samplerate=None, **kwargs):
        subtype = kwargs.get("subtype", subtype_of_wf(self.wf))
        samplerate = samplerate or self.sr
        if isinstance(filepath, int):
            rand_range = filepath
            template = "sound_save_{:0" + str(int(ceil(log10(rand_range)))) + ".0}.wav"
            filepath = template.format(randint(0, rand_range))
        else:
            filepath = filepath or "sound_save.wav"
        sf.write(filepath, self.wf, samplerate=samplerate, subtype=subtype, **kwargs)

    ####################################################################################################################
    # TRANSFORMATIONS

    def ensure_mono(self):
        self.wf = ensure_mono(self.wf)

    def crop_with_idx(self, first_idx, last_idx):
        cropped_sound = self.copy()
        cropped_sound.wf = cropped_sound.wf[first_idx : (last_idx + 1)]
        return cropped_sound

    def crop_with_seconds(self, first_second, last_second):
        return self.crop_with_idx(
            int(round(first_second * self.sr)),
            int(round(last_second * self.sr)),
        )

    def melspectr_matrix(self, **mel_kwargs):
        """
        Returns a melspectrogram matrix to be used to display a melspectrogram
        """
        mel_kwargs = dict(
            {"n_fft": 2048, "hop_length": 512, "n_mels": 128}, **mel_kwargs
        )
        S = melspectrogram(array(self.wf).astype(float), sr=self.sr, **mel_kwargs)
        # Convert to log scale (dB). We'll use the peak power as reference.
        return amplitude_to_db(S, ref=max)

    def __add__(self, append_sound):
        assert (
            self.sr == append_sound.sr
        ), "Sounds need to have the same sample rate to be appended"
        return Sound(sr=self.sr, wf=hstack(self.wf, append_sound.wf))

    ####################################################################################################################
    # DISPLAY FUNCTIONS

    def hear(self, autoplay=False, **kwargs):
        """
        Display UI to play sound
        """
        wf = array(ensure_mono(self.wf)).astype(float)
        wf[
            randint(len(wf))
        ] *= 1.001  # hack to avoid having exactly the same sound twice (creates an Audio bug)
        return Audio(data=wf, rate=self.sr, autoplay=autoplay, **kwargs)

    def plot_wf(*args, **kwargs):
        wf, sr = wf_and_sr(*args, **kwargs)
        return plot_wf(wf, sr)

    def display(self, autoplay=False, **kwargs):
        """
        Display a melspectrogram of sound and UI to play sound
        """
        self.melspectrogram(plot_it=True, **kwargs)

        return self.hear(autoplay=autoplay)

    def melspectrogram(self, plot_it=False, **mel_kwargs):
        """
        Returns a melsepectrogram matrix and plots a melspectrogram if plot_it is True
        """
        mel_kwargs = dict(
            {"n_fft": 2048, "hop_length": 512, "n_mels": 128}, **mel_kwargs
        )
        log_S = self.melspectr_matrix(**mel_kwargs)
        if plot_it:
            plot_melspectrogram(log_S, sr=self.sr, hop_length=mel_kwargs["hop_length"])
        return log_S

    def _repr_html_(self):
        return self.display()



"""
Generate simple waveforms for audio synthesis and testing.

Provides functions to create random samples, pure tones, triangular tones,
square tones, and annotated waveforms with different characteristics.
"""

import numpy as np
import random
import itertools
from typing import Mapping, Callable, Sequence, Any, Optional, Iterable, TypeVar
from collections import defaultdict

DFLT_SR = 44100
DFLT_CHK_SIZE = 21 * 2048
DFLT_MAX_AMPLITUDE = 30000
DFLT_PATTERN_LEN = 100
DFLT_PATTERN = [DFLT_MAX_AMPLITUDE] * 10 + [-DFLT_MAX_AMPLITUDE] * 10


def mk_audio_store(chk_size=DFLT_CHK_SIZE):
    return {
        "rand_1": random_samples(chk_size, max_amplitude=30000 / 3),
        "rand_2": random_samples(chk_size, max_amplitude=30000 / 2),
        "rand_3": random_samples(chk_size, max_amplitude=30000 / 1),
        "pure_1": pure_tone(chk_size, freq=440),
        "pure_2": pure_tone(chk_size, freq=440 * 1.4),
        "pure_3": pure_tone(chk_size, freq=440 * 1.8),
        "triangle_1": triangular_tone(chk_size, freq=440),
        "triangle_2": triangular_tone(chk_size, freq=440 * 1.4),
        "triangle_3": triangular_tone(chk_size, freq=440 * 1.8),
        "square_1": square_tone(chk_size, freq=440),
        "square_2": square_tone(chk_size, freq=440 * 1.4),
        "square_3": square_tone(chk_size, freq=440 * 1.4),
    }


def chk_from_pattern(chk_size=DFLT_CHK_SIZE, pattern=None):
    """
    Returns a chk with length chk_size that repeats pattern if given, or creates a random pattern of length 100
    >>> np.random.seed(1)
    >>> assert all(chk_from_pattern(5) == [3003, -17828, -24808, 2511, 20057])
    >>> chk = chk_from_pattern(6, [1,2,3])
    >>> assert all(chk == [1, 2, 3, 1, 2, 3])
    """
    if pattern is None:
        pattern = random_samples(DFLT_PATTERN_LEN, DFLT_MAX_AMPLITUDE)
    return np.tile(pattern, reps=int(np.ceil(chk_size / float(len(pattern)))))[
        :chk_size
    ].astype(np.int16)


def random_samples(chk_size=DFLT_CHK_SIZE, max_amplitude=DFLT_MAX_AMPLITUDE, **kwargs):
    """
    Returns a random sample of integers of length chk_size in the range [-max_amplitude, max_amplitude]
    >>> np.random.seed(1)
    >>> assert all(random_samples(5) == [3003, -17828, -24808, 2511, 20057])
    """
    return np.random.randint(-max_amplitude, max_amplitude, chk_size).astype(np.int16)


def pure_tone(
    chk_size=DFLT_CHK_SIZE,
    freq=440,
    sr=DFLT_SR,
    max_amplitude=DFLT_MAX_AMPLITUDE,
):
    """
    Generates a pure tone using given arguments
    >>> np.random.seed(1)
    >>> assert all(pure_tone(5) == [0, 1902, 3797, 5677, 7534])
    """
    pattern_length = int(sr / freq)
    pattern = max_amplitude * np.sin(np.linspace(0, 2 * np.pi, pattern_length))
    return chk_from_pattern(chk_size, pattern)


def triangular_tone(
    chk_size=DFLT_CHK_SIZE,
    freq=440,
    sr=DFLT_SR,
    max_amplitude=DFLT_MAX_AMPLITUDE,
):
    """
    Generates a triangular tone using given arguments
    >>> np.random.seed(1)
    >>> assert all(triangular_tone(5) == [-30000, -29900, -29800, -29700, -29600])
    """
    pattern_length = int(sr / freq)
    pattern = np.arange(-max_amplitude, max_amplitude, pattern_length)
    return chk_from_pattern(chk_size, pattern)


def square_tone(
    chk_size=DFLT_CHK_SIZE,
    freq=440,
    sr=DFLT_SR,
    max_amplitude=DFLT_MAX_AMPLITUDE,
):
    """
    Generates a square tone using given arguments
    >>> np.random.seed(1)
    >>> assert all(square_tone(5) == [30000, 30000, 30000, 30000, 30000])
    """
    pattern_length = int(sr / freq)
    half_pattern_length = int(pattern_length / 2)  # oh well for the plus minus 1
    pattern = [max_amplitude] * half_pattern_length + [
        -max_amplitude
    ] * half_pattern_length
    return chk_from_pattern(chk_size, pattern)


tag_to_wf_gen_func = {
    "random": random_samples,
    "pure_tone": pure_tone,
    "triangular_tone": triangular_tone,
    "square_tone": square_tone,
}

tag_to_wf_gen_func_items = tuple(tag_to_wf_gen_func.items())


class AnnotatedWaveform(object):
    """
    Creates a waveform with chunks corresponding to the tags in tag_to_wf_gen_func.
    >>> np.random.seed(1)
    >>> annotated_wf = AnnotatedWaveform(chk_size = 4)
    >>> chk_tag_gen = annotated_wf.chk_and_tag_gen()
    >>> wf, annots = annotated_wf.get_wf_and_annots()
    >>> assert list(wf) == [
    ...     3003, -17828, -24808, 2511, 0, 1902, 3797, 5677, -30000,
    ...     -29900, -29800, -29700, 30000, 30000, 30000, 30000
    ... ]
    >>> assert list(next(chk_tag_gen)[0]) == [20057, 13723, -22187, 22047]
    """

    def __init__(
        self,
        chk_size=DFLT_CHK_SIZE,
        freq=440,
        sr=DFLT_SR,
        max_amplitude=DFLT_MAX_AMPLITUDE,
    ):
        self.chk_size = chk_size
        self.sr = sr
        self.freq = freq
        self.max_amplitude = max_amplitude
        self._default_kwargs = {
            "chk_size": chk_size,
            "freq": freq,
            "sr": sr,
            "max_amplitude": max_amplitude,
        }

    def chk_and_tag_gen(
        self,
        chk_tags=("random", "pure_tone", "triangular_tone", "square_tone"),
    ):
        """
        Yields (chk, tag) pairs for each tag given in chk_tags
        """
        for tag in chk_tags:
            yield tag_to_wf_gen_func[tag](**self._default_kwargs), tag

    def get_wf_and_annots(
        self,
        chk_tags=("random", "pure_tone", "triangular_tone", "square_tone"),
    ):
        """
        Yields (wf, annots) tuple where annots is a dictionary mapping tag to chunk indices
        """
        slice_of_tag = defaultdict(list)
        bt_cursor = 0
        current_tag = None
        wf = list()
        for chk, tag in self.chk_and_tag_gen(chk_tags):
            wf += list(chk)
            if tag == current_tag:
                # TODO: Shame!
                slice_of_tag[tag][-1] = list(slice_of_tag[tag][-1])
                slice_of_tag[tag][-1][1] += self.chk_size
                slice_of_tag[tag][-1] = tuple(slice_of_tag[tag][-1])
            else:
                current_tag = tag
                slice_of_tag[tag].append((bt_cursor, bt_cursor + self.chk_size))

            bt_cursor += self.chk_size

        return np.array(wf), dict(slice_of_tag)


Tag = TypeVar("Tag")  # TODO: Use generic for better typing


def tag_wf_gen(
    tag_wfgen_map: Mapping[Tag, Callable[..., Sequence]] = tag_to_wf_gen_func_items,
    tag_sequence: Optional[Iterable[Tag]] = None,
):
    """Generate tagged waveforms -- i.e. ``(tag, wf)`` pairs.

    :param tag_wfgen_map: A ``{tag: wfgen, ...}`` map where wfgen is a callable taking
        no arguments and returning a sequence
    :param tag_sequence: A sequence of tags (that should all be keys of tag_wfgen_map)
    :return:

    Make a generator (here with default specs):

    >>> gen = tag_wf_gen()

    From it, source ``(tag, wf)`` pairs (by default, indefinitely).

    >>> tag, wf = next(gen)

    Since by default, we're given random ``tag`` and ``wf``, the only thing we can
    verify here is that tag is a string and ``wf`` has the default size of 43008.

    >>> type(tag)
    <class 'str'>
    >>> len(wf)
    43008

    By default the tags are randomly chosen from the ``tag_wfgen_map``, but you can
    also specify a iterable of tags explicitly. Obviously, the elements of this
    sequence must actually be keys of ``tag_wfgen_map`` -- this will be asserted if
    the iterable is a Sequence.

    You can specify a sequence of tags:

    >>> gen = tag_wf_gen(
    ...     tag_sequence=['random', 'pure_tone', 'triangular_tone', 'square_tone']
    ... )
    >>> tag_and_wf_seq = list(gen)  # consume the whole gen
    >>> [x[0] for x in tag_and_wf_seq]
    ['random', 'pure_tone', 'triangular_tone', 'square_tone']

    The first 5 numbers of the triangular tone

    >>> assert list(tag_and_wf_seq[2][1][:5]) == [-30000, -29900, -29800, -29700, -29600]

    """
    tag_wfgen_map = dict(tag_wfgen_map)

    if tag_sequence is None:
        indefinite_random_choice_from_tags = map(
            random.choice, itertools.repeat(list(tag_wfgen_map))
        )
        tag_sequence = indefinite_random_choice_from_tags
    else:
        if isinstance(tag_sequence, str):
            tag_sequence = tag_sequence.split()
        if isinstance(tag_sequence, Sequence):
            tag_sequence = list(tag_sequence)
            # TODO: Make it so that any tags can be mapped to a unique wf generator
            if not all(tag in tag_wfgen_map for tag in tag_sequence):
                raise ValueError(
                    f"A tag_sequence must only have elements that are keys of "
                    f"tag_wfgen_map, i.e. from {list(tag_wfgen_map)}"
                )

    for tag in tag_sequence:
        yield tag, tag_wfgen_map[tag]()



"""
Infinite waveform generation tools.

Provides classes and functions to create and manipulate waveforms that can be
queried for arbitrary time segments.
"""

import random
import numpy as np


class InfiniteWaveform(object):
    """
    A tiny little class emulating an infinite waveform with a given spectrum.
    Note that if choosing to add noise to the waveform, the generator will be considerably slower
    """

    def __init__(self, gen_spectrum, noise_amp=None):
        """
        noise_amp is the amplitude of the random noise added to the waveform,
        proportionally to the range of amplitude in the signal
        """

        self.reconstituted_wf = np.fft.irfft(gen_spectrum)
        self.noise_amp = noise_amp
        if noise_amp:
            max_signal = np.max(self.reconstituted_wf)
            min_signal = np.min(self.reconstituted_wf)
            max_amp = max_signal - min_signal
            self.noise_amp = max_amp * noise_amp
        else:
            self.noise_amp = 0
        self.win_size = len(self.reconstituted_wf)

    def query(self, bt, tt):
        """
        Returns a generator of the infinite waveform from bt to tt
        """
        if not self.noise_amp:
            for i in range(bt, tt):
                yield self.reconstituted_wf[i % self.win_size]
        else:
            # a random seed is chosen for each i, this allows for consistency between queries
            # the drawback is that it slows down the iterator quite a bit: 30 times slower for 1 millions samples
            for i in range(bt, tt):
                random.seed(a=i)
                noise = random.random() * self.noise_amp
                yield self.reconstituted_wf[i % self.win_size] + noise

    def __getitem__(self, idx):
        random.seed(a=idx)
        return (
            self.reconstituted_wf[idx % self.win_size]
            + random.random() * self.noise_amp
        )



"""
Generating Signals
"""

import numpy as np
import random
from itertools import chain
from typing import Iterable
import pandas as pd

DFLT_WORD_LENGTH = 30
DFLT_ALPHABET = "abcde"


def normal_dist(mu, sigma):
    """
    Returns a random sample from a normal distribution with mean mu and variance sigma
    """
    return sigma * np.random.randn() + mu


def gen_words(
    N=DFLT_WORD_LENGTH,
    alphabet: Iterable = DFLT_ALPHABET,
    spread_pct=0.01,
    proba_dist="normal",
):
    """
    Returns a generator of lists, with each consisting of n repetitions of a random word for alphabet, with n being
    determined by a normal distribution with mean N and variance N * spread_pct
    >>> gen = gen_words(N=3, alphabet=('foo',), spread_pct=0)
    >>> assert next(gen) == ['foo', 'foo', 'foo']
    """
    if proba_dist == "normal":
        sigma = N * spread_pct
        mu = N
        dist = normal_dist(mu, sigma)
    else:
        raise NotImplementedError(
            f"Probability distribution {proba_dist} not implemented"
        )
    while True:
        length = dist
        word = [random.choice(alphabet)] * int(length)
        yield word


def categorical_gen(gen_it):
    yield from chain.from_iterable(gen_it())


def alphabet_to_bins(alphabet=DFLT_ALPHABET):
    """
    Returns a dictionary matching each word in alphabet to bins of size 10.0 ranging from 0 to 10 * len(alphabet)
    >>> assert alphabet_to_bins() == {
    ...     'a': (0.0, 10.0), 'b': (10.0, 20.0), 'c': (20.0, 30.0),
    ...     'd': (30.0, 40.0), 'e': (40.0, 50.0)
    ... }
    """
    length = len(alphabet)
    low = 0.0
    high = 10.0 * length
    bins = np.linspace(low, high, length + 1)
    # print(bins)
    return {letter: (bins[i], bins[i + 1]) for i, letter in enumerate(alphabet)}


# def context_to_signal(cat_gen, dict_symbol_to_interval)
#   symbol = next(cat_gen)
#   low, hi = dict_symbol_to_interval[symbol]
#   res = np.random.uniform(low=0.0, high=1.0)


def call_repeatedly(func, *args, **kwargs):
    """
    Returns a generator that calls func repeatedly with the given args and kwargs
    >>> gen = call_repeatedly(alphabet_to_bins)
    >>> assert next(gen) == {'a': (0.0, 10.0), 'b': (10.0, 20.0), 'c': (20.0, 30.0), 'd': (30.0, 40.0), 'e': (40.0, 50.0)}
    >>> assert next(gen) == {
    ...     'a': (0.0, 10.0), 'b': (10.0, 20.0), 'c': (20.0, 30.0),
    ...     'd': (30.0, 40.0), 'e': (40.0, 50.0)
    ... }

    """
    while True:
        yield func(*args, **kwargs)


def bernoulli(p_out=0.1):
    """
    Returns a random sample of a bernoulli distribution with probability p_out
    >>> assert bernoulli(0) == 0
    >>> assert bernoulli(1) == 1
    """
    a = [0, 1]
    p = [1.0 - p_out, p_out]
    return int(np.random.choice(a, size=1, replace=True, p=p))


def bernoulli_gen(p_out=0.5):
    """
    Returns a generator that returns random samples of a bernoulli distribution with probability p_out
    """
    return call_repeatedly(bernoulli, p_out=p_out)


def inlier_outlier(segment, interval_size, outlier_status):
    """
    >>> assert inlier_outlier([0,1], 10, 0) < 1
    >>> low = np.random.randint(5, 10)
    >>> high = np.random.randint(low, 20)
    >>> assert inlier_outlier([low, high], low*high, 1) < low + low*high
    """
    low, high = segment
    if outlier_status == 0:
        return np.random.uniform(low, high)
    else:
        low, high = high, low + interval_size
        return np.random.uniform(low, high) % interval_size


def signal(symbol_gen, outlier_gen, alphabet):
    while True:
        symb = next(symbol_gen)
        outlier = next(outlier_gen)
        length = len(alphabet)
        low, high = alphabet_to_bins(alphabet)[symb]
        yield inlier_outlier((low, high), length * 10, outlier)


def create_session(symbol_gen, outlier_gen, alphabet, session_length=50):
    symbs = []
    outliers = []
    sigs = []
    for _ in range(session_length):
        symb = next(symbol_gen)
        symbs.append(symb)
        outlier = next(outlier_gen)
        outliers.append(outlier)
        length = len(alphabet)
        low, high = alphabet_to_bins(alphabet)[symb]
        sigs.append(inlier_outlier((low, high), length * 10, outlier))
    return symbs, outliers, sigs


def string_to_num(word):
    """
    Converts a string to a list of numbers corresponding to the alphabetical order of the characters in the string
    >>> assert string_to_num('abc') == [0, 1, 2]
    >>> assert string_to_num('generator') == [2, 1, 3, 1, 5, 0, 6, 4, 5]
    """
    all_letters = sorted(list(set(word)))
    return [all_letters.index(letter) for letter in word]


def session_to_df(session):
    symbs, outliers, sigs = session
    df = pd.DataFrame()
    df["symbols"] = string_to_num("".join(symbs))
    df["outliers"] = outliers
    df["signal"] = sigs
    return df



"""
Voice-based time announcement utilities.

Provides functions and classes to generate spoken time announcements at
regular intervals with configurable voices and formats.
"""

from datetime import datetime as dt
import time
from contextlib import suppress

with suppress(ModuleNotFoundError, ImportError):
    import pyttsx3

    DFLT_TIME_FORMAT = "%H %M %S"

    class Voicer:
        def __init__(
            self,
            voice=None,
            volume=None,
            rate=None,
            engine_kwargs=None,
            time_format=DFLT_TIME_FORMAT,
            **kwargs
        ):
            engine_kwargs = engine_kwargs or {}
            self.engine = pyttsx3.init(**engine_kwargs)  # object creation

            if voice is None:
                for voice_dict in self.voices:
                    if "en_US" in voice_dict["languages"]:
                        voice = voice_dict["id"]
                        break
                else:
                    print("Didn't find a voice!!")
            else:
                voice_id = self.voice_id_for_name(voice)
                if voice_id is not None:
                    voice = voice_id

            for attr, val in dict(kwargs, volume=volume, rate=rate).items():
                if val is not None:
                    self.engine.setProperty(attr, val)

            self.time_format = time_format

        def say(self, text):
            self.engine.say(text)
            self.engine.runAndWait()
            self.engine.stop()

        @property
        def voices_df(self):
            import pandas as pd

            df = pd.DataFrame(self.voices)
            df = df.set_index("name")
            df["language"] = [x[0] for x in df["languages"]]
            df["gender"] = [
                "male" if x == "VoiceGenderMale" else "female" for x in df["gender"]
            ]
            del df["languages"]
            return df

        def __getattribute__(self, attr):
            if attr in {"volume", "rate", "voice"}:
                return self.engine.getProperty(attr)
            else:
                return super().__getattribute__(attr)

        @property
        def voices(self):
            return [x.__dict__ for x in self.engine.getProperty("voices")]

        def voice_id_for_name(self, name):
            for voice in self.voices:
                if voice["name"] == name:
                    return voice["id"]

        def say_the_time(self, verbose=False):
            s = dt.utcnow().strftime(self.time_format)
            if verbose:
                print(s)
            self.say(s)

        def tell_time_continuously(self, every_secs=5, verbose=False):
            try:
                while True:
                    tic = time.time()
                    self.say_the_time(verbose=verbose)
                    elapsed = time.time() - tic
                    time.sleep(max(0, every_secs - elapsed))
            except KeyboardInterrupt:
                print("KeyboardInterrupt!!!")
                self.engine.stop()

    def tell_time_continuously(
        every_secs=3,
        voice=None,
        volume=1,
        rate=200,
        engine_kwargs=None,
        time_format=DFLT_TIME_FORMAT,
        verbose=False,
    ):
        """
        Loop, and say the time regularly.

        Args:
            every_secs: How often to tell the time (in seconds)
            voice: What voice to use (or "help" to get a list of names (and their properties))
            volume: Volume of voice
            rate: How quickly the time is uttered
            engine_kwargs: Stuff you don't need to worry about
            time_format: In what format the time (or even date) is transformed to text that is read.
                See https://www.programiz.com/python-programming/datetime/strftime
            verbose: Print the time every loop

        Returns:
            Nothing, just loops and says the time regularly
        """
        if voice == "help":
            print(Voicer().voices_df)
        else:
            voicer = Voicer(
                voice=voice,
                volume=volume,
                rate=rate,
                engine_kwargs=engine_kwargs,
                time_format=time_format,
            )
            voicer.tell_time_continuously(every_secs, verbose=verbose)

    if __name__ == "__main__":
        import argh

        argh.dispatch_command(tell_time_continuously)



"""
Tools for generating diagnostic audio signals.

Provides classes and functions to create special sound patterns useful for
testing and diagnosing audio processing pipelines and equipment.
"""

from numpy import (
    array,
    hstack,
    ones,
    ceil,
    zeros,
    floor,
    argmin,
    diff,
    where,
    reshape,
    arange,
    iinfo,
    int16,
    linspace,
    pi,
    unique,
    tile,
    repeat,
    all,
)
import numpy as np
import math
from typing import Optional
from numpy.random import randint
from itertools import islice, count
from datetime import datetime as dt

from loopyng.gen.util import DFLT_N_SAMPLES, DFLT_SR, DFLT_FREQ

second_ms = 1000.0
epoch = dt.fromtimestamp(0)


def utcnow_ms():
    """
    Returns the current time in UTC in milliseconds
    """
    return (dt.utcnow() - epoch).total_seconds() * second_ms


def window(seq, n=2):
    """
    Returns a sliding window (of width n) over data from the iterable
    s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...
    """
    it = iter(seq)
    result = tuple(islice(it, n))
    if len(result) == n:
        yield result
    for elem in it:
        result = result[1:] + (elem,)
        yield result


def ums_to_01_array(ums, n_ums_bits):
    """
    Converts ums to an array with length n_ums_bits equivalent to a binary representation of ums
    assert all(ums_to_01_array(100,1) == [1, 1, 0, 0, 1, 0, 0])
    """
    ums_bits_str_format = "{:0" + str(n_ums_bits) + "b}"
    return array([int(x == "1") for x in ums_bits_str_format.format(ums)])


class BinarySound(object):
    def __init__(
        self,
        redundancy,
        repetition,
        nbits=50,
        header_size_words=1,
        header_pattern=None,
    ):
        """
        :param redundancy: num of times a word will be repeated (to implement error correction)
        :param repetition: num of times to repeat each element of a pattern.
        :param nbits: num of bits of a pattern. An {0,1}-array to be encoded will have to be this size.
            Also, if a header_pattern is explicitly given, it will also have to be this size.
        :param header_size_words: the size (in num of words) of the header
        :param header_pattern: specifies the header pattern. This pattern will be repeated (i.e. it's elements
            repeated several times (010-->000011110000) and tiled (header_size_words times). Can be:
            * None: Default. Will choose a random array of 0s and 1s
            * an explicit array of nbits 0s and 1s to use for the header
            * a string indicating a method to generate it (for now, choices are "halfhalf" or "alternating")

        # TODO: Repair: TypeError: unsupported operand type(s) for *: 'int' and 'map'

        >>> from numpy import *
        >>> from numpy.random import randint
        >>>
        >>> nbits=50
        >>> bs = BinarySound(
        ...     nbits=nbits, redundancy=142, repetition=3, header_size_words=1)
        >>> utc = randint(0, 2, nbits)
        >>> wf = bs.mk_phrase(utc)
        >>> print(bs)
        {'repetition': 3, 'word_size_frm': 150, 'redundancy': 142, 'phrase_data_frm': 21300}
        >>> assert all(utc == bs.decode(wf))

        """
        self.nbits = nbits

        self.header_size_words = header_size_words
        # repetition: how many times to repeat each bit to make a word
        self.repetition = repetition
        # word_size_frm: the size of a word, in frames
        self.word_size_frm = int(self.nbits * self.repetition)
        # redundancy: how many times to repeat a word to make (along with header) a phrase
        self.redundancy = redundancy
        if header_pattern is None:
            header_pattern = randint(0, 2, nbits)
        elif isinstance(header_pattern, str):
            if header_pattern == "halfhalf":
                header_pattern = hstack(
                    (ones(int(ceil(nbits / 2))), zeros(int(floor(nbits / 2))))
                ).astype(int)
            elif header_pattern == "alternating":
                header_pattern = array([1, 0] * int(ceil(nbits / 2)))[:nbits]
            else:
                raise ValueError(
                    "header_pattern not recognized: {}".format(header_pattern)
                )
        else:
            assert (
                len(header_pattern) == nbits
            ), "header_pattern must have nbits={}".format(nbits)
            assert set(unique(header_pattern).astype(int)) == {
                0,
                1,
            }, "header_pattern must be an array of 0s and 1s"
        self.header_pattern = header_pattern
        self.header_word = tile(
            repeat(self.header_pattern, self.repetition),
            self.header_size_words,
        )

        self.phrase_data_frm = self.redundancy * self.word_size_frm

    @classmethod
    def for_audio_params(
        cls,
        nbits=50,
        freq=3000,
        chk_size=43008,
        sr=DFLT_SR,
        header_size_words=1,
        header_pattern=None,
    ):
        """
        Construct a BinarySound object for a set of audio params
        :param nbits: num of bits of a word of data we want to encode
        :param freq: frequency (num of times per second the bits will be repeated -- with sr, will determine repetition)
        :param chk_size: chunk size (in frames) of the sounds we'll be using (determines
        :param sr: sample rate of the targeted sound
        :param header_size_words: header size (increasing it will decrease error rate, but increase computation time)
        :param header_pattern: specifies the header pattern. This pattern will be repeated (i.e. it's elements
            repeated several times (010-->000011110000) and tiled (header_size_words times). Can be:
            * None: Default. Will choose a random array of 0s and 1s
            * an explicit array of nbits 0s and 1s to use for the header
            * a string indicating a method to generate it (for now, choices are "halfhalf" or "alternating")
        :return: a BinarySound object
        >>> from numpy import *
        >>> from numpy.random import randint
        >>>
        >>> nbits=50
        >>> bs = BinarySound.for_audio_params(
        ...     nbits=nbits, freq=6000, chk_size=43008, sr=44100, header_size_words=1)
        >>> utc = randint(0, 2, nbits)
        >>> wf = bs.mk_phrase(utc)
        >>> print(bs)
        {'repetition': 3, 'word_size_frm': 150, 'redundancy': 142, 'phrase_data_frm': 21300}
        >>> assert all(utc == bs.decode(wf))

        """
        # repetition: how many times to repeat each bit to make a word
        repetition = int(floor(sr / (2 * freq)))
        # word_size_frm: the size of a word, in frames
        word_size_frm = int(nbits * repetition)
        # redundancy: how many times to repeat a word to make (along with header) a phrase
        redundancy = int(floor((chk_size / 2) / word_size_frm) - header_size_words)

        self = cls(
            nbits=nbits,
            redundancy=redundancy,
            repetition=repetition,
            header_size_words=header_size_words,
            header_pattern=header_pattern,
        )
        self.freq = freq
        self.sr = sr
        self.chk_size = chk_size
        self.redundancy = redundancy
        self.repetition = repetition
        return self

    def mk_phrase(self, bit_array):
        repeated_bit_array = repeat(bit_array, self.repetition)
        tiled = tile(repeated_bit_array, self.redundancy)
        wf = hstack((self.header_word, tiled))
        return (2 * wf) - 1

    def mk_utc_phrases(self, sound_duration_s=12):
        wf = list()

        def wf_seconds(wf):
            return len(wf) / float(self.sr)

        while wf_seconds(wf) < sound_duration_s:
            bit_array = ums_to_01_array(ums=int(utcnow_ms()), n_ums_bits=self.nbits)
            wf += list(self.mk_phrase(bit_array))

        return array(wf)[: int(sound_duration_s * self.sr)]

    def header_position(self, wf):
        wf = wf > 0
        return argmin(slow_mask(wf, self.header_word))

    def decode(self, wf):
        header_pos = self.header_position(wf)
        header_end_idx = header_pos + len(self.header_word)
        wf = wf > 0
        m = wf[header_end_idx : (header_end_idx + self.phrase_data_frm)]
        size_to_make_it_a_multiple_of_word_size = len(m) - (len(m) % self.word_size_frm)
        m = m[:size_to_make_it_a_multiple_of_word_size]
        m = reshape(m, (-1, self.word_size_frm))
        m = m.sum(axis=0).reshape((-1, self.repetition))
        m = m.sum(axis=1)
        return (m / float(self.repetition * self.redundancy) > 0.5).astype(int)

    def __repr__(self):
        return str(
            {
                "repetition": self.repetition,
                "word_size_frm": self.word_size_frm,
                "redundancy": self.redundancy,
                "phrase_data_frm": self.phrase_data_frm,
            }
        )


def zero_crossing_gaps(wf):
    w = wf > 0
    ww = hstack((w, ~w[-1]))
    return diff(hstack(([0], 1 + where(diff(ww))[0])))


def slow_mask(arr, msk):
    msk = array(msk)
    arr_msk_dist = list()
    for w in window(arr, len(msk)):
        arr_msk_dist.append(sum(abs(array(w) - msk)))
    return arr_msk_dist


class WfGen(object):
    """
    >>> wfgen = WfGen(sr=44100, buf_size_frm=2048, amplitude=0.5)
    >>> lookup = wfgen.mk_lookup_table(freq=4400)
    >>> assert len(lookup) == 10
    >>> wfgen.mk_sine_wf(n_frm=5, freq=4400)
    array([0.        , 0.293316  , 0.47508605, 0.47618432, 0.29619315])
    """

    def __init__(self, sr=DFLT_SR, buf_size_frm=2048, amplitude=0.5):
        self.sr = sr
        self.buf_size_frm = buf_size_frm
        self.buf_size_s = buf_size_frm / float(self.sr)
        if amplitude > 1.0:
            amplitude = 1.0
        if amplitude < 0.0:
            amplitude = 0.0
        self.amplitude = float(amplitude)
        self.lookup_table_freqs = (arange(int(buf_size_frm / 2)) + 1) / self.buf_size_s
        self.lookup_tables = list(map(self.mk_lookup_table, self.lookup_table_freqs))

    def mk_sine_wave_from_lookup_table(self, lookup_table):
        period = len(lookup_table)
        return (lookup_table[i % period] for i in count(0))

    def mk_sine_wave_iterator(self, freq=440):
        if isinstance(freq, (int, float)):
            table_idx = where(self.lookup_table_freqs == freq)[0]
            if len(table_idx) > 0:
                table_idx = table_idx[0]
                lookup_table = self.lookup_tables[table_idx]
            else:
                lookup_table = self.mk_lookup_table(freq=freq)
            return self.mk_sine_wave_from_lookup_table(lookup_table)
        else:  # consider freq to already be a lookup table
            return self.mk_sine_wave_from_lookup_table(freq)

    def mk_sine_wf(self, n_frm, freq=440):
        it = self.mk_sine_wave_iterator(freq)
        return array([x for x in islice(it, int(n_frm))])

    def mk_lookup_table(self, freq=440):
        freq = float(freq)
        period = int(self.sr / freq)
        lookup_table = [
            self.amplitude
            * math.sin(
                2.0 * math.pi * float(freq) * (float(i % period) / float(self.sr))
            )
            for i in range(period)
        ]
        return lookup_table

    def mk_wf_from_freq_weight_array(self, n_frm, freq_weight_array):
        wf = zeros(n_frm)
        for i, w in enumerate(freq_weight_array):
            wf += w * self.mk_sine_wf(n_frm, self.lookup_tables[i])
        return wf


class TimeSound(WfGen):
    def __init__(self, sr=DFLT_SR, buf_size_frm=2048, amplitude=0.5, n_ums_bits=30):
        super(TimeSound, self).__init__(
            sr=sr, buf_size_frm=buf_size_frm, amplitude=amplitude
        )
        self.n_ums_bits = n_ums_bits
        self.ums_bits_str_format = "{:0" + str(n_ums_bits) + "b}"
        self.n_freqs_per_ums_bit = len(self.lookup_tables) // self.n_ums_bits
        self.n_freqs_for_ums = self.n_freqs_per_ums_bit * self.n_ums_bits
        self.buf_size_ms = self.buf_size_s * 1000

    def ums_to_01_array(self, ums):
        return array([int(x == "1") for x in self.ums_bits_str_format.format(ums)])

    def freq_weight_array_for_ums(self, ums):
        return tile(self.ums_to_01_array(ums), self.n_freqs_per_ums_bit)

    def ums_to_wf(self, ums, n_bufs=1):
        return self.mk_wf_from_freq_weight_array(
            n_frm=n_bufs * self.buf_size_frm,
            freq_weight_array=self.freq_weight_array_for_ums(ums),
        )

    def timestamped_wf(self, offset_ums=0, n_bufs=21, n_bufs_per_tick=1):
        wf = list()
        ums = offset_ums
        for buf_idx in range(n_bufs):
            wf.extend(list(self.ums_to_wf(ums, n_bufs=n_bufs_per_tick)))
            ums = int(ums + n_bufs_per_tick * self.buf_size_ms)
        return array(wf)

    def spectr_of_time(self, offset_ums=0, n_bufs=21, n_bufs_per_tick=1):
        wf = list()
        ums = offset_ums
        for buf_idx in range(n_bufs):
            wf.append(list(self.freq_weight_array_for_ums(ums)))
            if n_bufs_per_tick > 1:
                wf.append(list(zeros(self.n_freqs_for_ums)))
            ums = int(ums + n_bufs_per_tick * self.buf_size_ms)
        wf = array(wf)
        return hstack(
            (
                wf,
                zeros((wf.shape[0], int(self.buf_size_frm / 2 - wf.shape[1]))),
            )
        )


import soundfile as sf

DFLT_BLEEP_LOC = 400
DFLT_BLEEP_SPEC = 100


def mk_some_buzz_wf(
    freq: float = DFLT_FREQ,
    n_samples: int = DFLT_N_SAMPLES,
    sr: float = DFLT_SR,
):
    """Produce a sawtooth waveform with given frequency and n_samples.
    The sample rate ``sr`` serves to interpret the ``freq`` specification in the number
    of samples unit.

    .. seealso:: right_triangles

    """
    return right_triangles(samples_per_period=sr / freq, n_samples=n_samples)


def right_triangles(samples_per_period: float = 3, n_samples: int = 7):
    """Simplified sawtooth waveform.

    >>> import numpy as np
    >>> assert np.all(
    ... right_triangles(samples_per_period=3, n_samples=7)
    ... == np.array(
    ...     [-32767, -10922,  10922, -32767, -10922,  10922, -32767],
    ...     dtype=int16
    ... ))
    """
    #     samples_per_period = sr / freq
    wf = np.linspace(0, n_samples - 1, n_samples)
    wf = np.mod(wf, samples_per_period)
    wf = 2 * (wf / samples_per_period) - 1

    return (wf * iinfo(int16).max).astype(int16)


def wf_with_timed_bleeps(
    n_samples: int = DFLT_SR * 2,
    bleep_loc=DFLT_BLEEP_LOC,
    bleep_spec=DFLT_BLEEP_SPEC,
    sr: float = DFLT_SR,
):
    """Not sure this works as expected. Docs needed."""
    if isinstance(bleep_spec, (int, float)):
        bleep_size = int(bleep_spec)
        bleep_spec = mk_some_buzz_wf(sr)[:bleep_size]
    if isinstance(bleep_loc, (int, float)):
        bleep_loc = range(0, n_samples, bleep_loc)

    bleep_size = len(bleep_spec)
    # bleep_loc = (sr * (array(bleep_loc_ms) / 1000)).astype(int)
    # max_bleep_loc = max(bleep_size) + len(bleep_spec)
    wf = zeros(n_samples)
    for loc_frm in bleep_loc:
        wf[loc_frm : (loc_frm + bleep_size)] = bleep_spec
    return wf


def mk_sounds_with_timed_bleeps(
    n_samples=DFLT_SR * 2,
    bleep_loc=DFLT_BLEEP_LOC,
    bleep_spec=DFLT_BLEEP_SPEC,
    sr=DFLT_SR,
    save_filepath="bleeps.wav",
):
    wf = wf_with_timed_bleeps(
        n_samples=n_samples, bleep_loc=bleep_loc, bleep_spec=bleep_spec, sr=sr
    )
    sf.write(save_filepath, data=wf, samplerate=sr, format="WAV")



"""A few utils (specshow, melspectrogram) vendored from librosa.

This code was copied from parts of librosa, and adapted, so as to be able to use
targeted functionality with less dependencies and manual installation
(namely for libsndfile) than librosa has.

Librosa can be found here: https://librosa.org/

Librosa's license follows.

ISC License
Copyright (c) 2013--2017, librosa development team.

Permission to use, copy, modify, and/or distribute this software for any purpose with or
without fee is hereby granted, provided that the above copyright notice and this
permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""

import warnings
import numpy as np
from matplotlib.cm import get_cmap
from matplotlib.axes import Axes
from matplotlib.ticker import Formatter, ScalarFormatter
from matplotlib.ticker import MaxNLocator
from matplotlib.ticker import SymmetricalLogLocator
import matplotlib
from packaging.version import parse as version_parse
import scipy
import scipy.signal
from numpy.lib.stride_tricks import as_strided
import re

MAX_MEM_BLOCK = 2**8 * 2**10

# specshow


def specshow(
    data,
    x_coords=None,
    y_coords=None,
    x_axis=None,
    y_axis=None,
    sr=22050,
    hop_length=512,
    fmin=None,
    fmax=None,
    tuning=0.0,
    bins_per_octave=12,
    key="C:maj",
    Sa=None,
    mela=None,
    thaat=None,
    auto_aspect=True,
    htk=False,
    ax=None,
    **kwargs,
):
    if np.issubdtype(data.dtype, np.complexfloating):
        warnings.warn(
            "Trying to display complex-valued input. " "Showing magnitude instead."
        )
        data = np.abs(data)

    kwargs.setdefault("cmap", cmap(data))
    kwargs.setdefault("rasterized", True)
    kwargs.setdefault("edgecolors", "None")
    kwargs.setdefault("shading", "flat")

    all_params = dict(
        kwargs=kwargs,
        sr=sr,
        fmin=fmin,
        fmax=fmax,
        tuning=tuning,
        bins_per_octave=bins_per_octave,
        hop_length=hop_length,
        key=key,
        htk=htk,
    )

    # Get the x and y coordinates
    y_coords = __mesh_coords(y_axis, y_coords, data.shape[0], **all_params)
    x_coords = __mesh_coords(x_axis, x_coords, data.shape[1], **all_params)

    axes = __check_axes(ax)

    out = axes.pcolormesh(x_coords, y_coords, data, **kwargs)
    __set_current_image(ax, out)

    axes.set_xlim(x_coords.min(), x_coords.max())
    axes.set_ylim(y_coords.min(), y_coords.max())

    # Set up axis scaling
    __scale_axes(axes, x_axis, "x")
    __scale_axes(axes, y_axis, "y")

    # Construct tickers and locators
    __decorate_axis(axes.xaxis, x_axis, key=key, Sa=Sa, mela=mela, thaat=thaat)
    __decorate_axis(axes.yaxis, y_axis, key=key, Sa=Sa, mela=mela, thaat=thaat)

    # If the plot is a self-similarity/covariance etc. plot, square it
    if __same_axes(x_axis, y_axis, axes.get_xlim(), axes.get_ylim()) and auto_aspect:
        axes.set_aspect("equal")
    return out


def cmap(data, robust=True, cmap_seq="magma", cmap_bool="gray_r", cmap_div="coolwarm"):
    data = np.atleast_1d(data)

    if data.dtype == "bool":
        return get_cmap(cmap_bool, lut=2)

    data = data[np.isfinite(data)]

    if robust:
        min_p, max_p = 2, 98
    else:
        min_p, max_p = 0, 100

    min_val, max_val = np.percentile(data, [min_p, max_p])

    if min_val >= 0 or max_val <= 0:
        return get_cmap(cmap_seq)

    return get_cmap(cmap_div)


def __mesh_coords(ax_type, coords, n, **kwargs):
    """Compute axis coordinates"""

    if coords is not None:
        if len(coords) < n:
            raise Exception(
                "Coordinate shape mismatch: " "{}<{}".format(len(coords), n)
            )
        return coords

    coord_map = {
        "linear": __coord_fft_hz,
        "fft": __coord_fft_hz,
        "fft_note": __coord_fft_hz,
        "fft_svara": __coord_fft_hz,
        "hz": __coord_fft_hz,
        "log": __coord_fft_hz,
        "mel": __coord_mel_hz,
        "cqt": __coord_cqt_hz,
        "cqt_hz": __coord_cqt_hz,
        "cqt_note": __coord_cqt_hz,
        "cqt_svara": __coord_cqt_hz,
        "chroma": __coord_chroma,
        "chroma_c": __coord_chroma,
        "chroma_h": __coord_chroma,
        "time": __coord_time,
        "s": __coord_time,
        "ms": __coord_time,
        "lag": __coord_time,
        "lag_s": __coord_time,
        "lag_ms": __coord_time,
        "tonnetz": __coord_n,
        "off": __coord_n,
        "tempo": __coord_tempo,
        "fourier_tempo": __coord_fourier_tempo,
        "frames": __coord_n,
        None: __coord_n,
    }

    if ax_type not in coord_map:
        raise Exception("Unknown axis type: {}".format(ax_type))
    return coord_map[ax_type](n, **kwargs)


def __coord_fourier_tempo(n, sr=22050, hop_length=512, **_kwargs):
    """Fourier tempogram coordinates"""

    n_fft = 2 * (n - 1)
    # The following code centers the FFT bins at their frequencies
    # and clips to the non-negative frequency range [0, nyquist]
    basis = fourier_tempo_frequencies(sr=sr, hop_length=hop_length, win_length=n_fft)
    fmax = basis[-1]
    basis -= 0.5 * (basis[1] - basis[0])
    basis = np.append(np.maximum(0, basis), [fmax])
    return basis


def fourier_tempo_frequencies(sr=22050, win_length=384, hop_length=512):
    # sr / hop_length gets the frame rate
    # multiplying by 60 turns frames / sec into frames / minute
    return fft_frequencies(sr=sr * 60 / float(hop_length), n_fft=win_length)


def frames_to_time(frames, sr=22050, hop_length=512, n_fft=None):
    samples = frames_to_samples(frames, hop_length=hop_length, n_fft=n_fft)

    return samples_to_time(samples, sr=sr)


def samples_to_time(samples, sr=22050):
    return np.asanyarray(samples) / float(sr)


def frames_to_samples(frames, hop_length=512, n_fft=None):
    offset = 0
    if n_fft is not None:
        offset = int(n_fft // 2)

    return (np.asanyarray(frames) * hop_length + offset).astype(int)


def __coord_time(n, sr=22050, hop_length=512, **_kwargs):
    """Get time coordinates from frames"""
    return frames_to_time(np.arange(n + 1), sr=sr, hop_length=hop_length)


def __coord_chroma(n, bins_per_octave=12, **_kwargs):
    """Get chroma bin numbers"""
    return np.linspace(0, (12.0 * n) / bins_per_octave, num=n + 1, endpoint=True)


def tempo_frequencies(n_bins, hop_length=512, sr=22050):
    bin_frequencies = np.zeros(int(n_bins), dtype=np.float)

    bin_frequencies[0] = np.inf
    bin_frequencies[1:] = 60.0 * sr / (hop_length * np.arange(1.0, n_bins))

    return bin_frequencies


def __coord_tempo(n, sr=22050, hop_length=512, **_kwargs):
    """Tempo coordinates"""
    basis = tempo_frequencies(n + 2, sr=sr, hop_length=hop_length)[1:]
    edges = np.arange(1, n + 2)
    return basis * (edges + 0.5) / edges


def __coord_cqt_hz(n, fmin=None, bins_per_octave=12, sr=22050, **_kwargs):
    """Get CQT bin frequencies"""
    if fmin is None:
        fmin = note_to_hz("C1")

    # Apply tuning correction
    fmin = fmin * 2.0 ** (_kwargs.get("tuning", 0.0) / bins_per_octave)

    # we drop by half a bin so that CQT bins are centered vertically
    freqs = cqt_frequencies(
        n + 1,
        fmin=fmin / 2.0 ** (0.5 / bins_per_octave),
        bins_per_octave=bins_per_octave,
    )

    if np.any(freqs > 0.5 * sr):
        warnings.warn(
            "Frequency axis exceeds Nyquist. "
            "Did you remember to set all spectrogram parameters in specshow?"
        )

    return freqs


def cqt_frequencies(n_bins, fmin, bins_per_octave=12, tuning=0.0):
    correction = 2.0 ** (float(tuning) / bins_per_octave)
    frequencies = 2.0 ** (np.arange(0, n_bins, dtype=float) / bins_per_octave)

    return correction * fmin * frequencies


def note_to_hz(note, **kwargs):
    return midi_to_hz(note_to_midi(note, **kwargs))


def midi_to_hz(notes):
    return 440.0 * (2.0 ** ((np.asanyarray(notes) - 69.0) / 12.0))


def note_to_midi(note, round_midi=True):
    if not isinstance(note, str):
        return np.array([note_to_midi(n, round_midi=round_midi) for n in note])

    pitch_map = {"C": 0, "D": 2, "E": 4, "F": 5, "G": 7, "A": 9, "B": 11}
    acc_map = {
        "#": 1,
        "": 0,
        "b": -1,
        "!": -1,
        "♯": 1,
        "𝄪": 2,
        "♭": -1,
        "𝄫": -2,
        "♮": 0,
    }

    match = re.match(
        r"^(?P<note>[A-Ga-g])"
        r"(?P<accidental>[#♯𝄪b!♭𝄫♮]*)"
        r"(?P<octave>[+-]?\d+)?"
        r"(?P<cents>[+-]\d+)?$",
        note,
    )
    if not match:
        raise Exception("Improper note format: {:s}".format(note))

    pitch = match.group("note").upper()
    offset = np.sum([acc_map[o] for o in match.group("accidental")])
    octave = match.group("octave")
    cents = match.group("cents")

    if not octave:
        octave = 0
    else:
        octave = int(octave)

    if not cents:
        cents = 0
    else:
        cents = int(cents) * 1e-2

    note_value = 12 * (octave + 1) + pitch_map[pitch] + offset + cents

    if round_midi:
        note_value = int(np.round(note_value))

    return note_value


def __coord_n(n, **_kwargs):
    """Get bare positions"""
    return np.arange(n + 1)


def __coord_mel_hz(n, fmin=0, fmax=None, sr=22050, htk=False, **_kwargs):
    """Get the frequencies for Mel bins"""

    if fmin is None:
        fmin = 0
    if fmax is None:
        fmax = 0.5 * sr

    basis = mel_frequencies(n, fmin=fmin, fmax=fmax, htk=htk)
    basis[1:] -= 0.5 * np.diff(basis)
    basis = np.append(np.maximum(0, basis), [fmax])
    return basis


def __coord_fft_hz(n, sr=22050, **_kwargs):
    """Get the frequencies for FFT bins"""
    n_fft = 2 * (n - 1)
    # The following code centers the FFT bins at their frequencies
    # and clips to the non-negative frequency range [0, nyquist]
    basis = fft_frequencies(sr=sr, n_fft=n_fft)
    fmax = basis[-1]
    basis -= 0.5 * (basis[1] - basis[0])
    basis = np.append(np.maximum(0, basis), [fmax])
    return basis


def __check_axes(axes):
    """Check if "axes" is an instance of an axis object. If not, use `gca`."""
    if axes is None:
        import matplotlib.pyplot as plt

        axes = plt.gca()
    elif not isinstance(axes, Axes):
        raise Exception(
            "`axes` must be an instance of matplotlib.axes.Axes. "
            "Found type(axes)={}".format(type(axes))
        )
    return axes


def __set_current_image(ax, img):
    """Helper to set the current image in pyplot mode.
    If the provided ``ax`` is not `None`, then we assume that the user is using the object API.
    In this case, the pyplot current image is not set.
    """

    if ax is None:
        import matplotlib.pyplot as plt

        plt.sci(img)


def __scale_axes(axes, ax_type, which):
    """Set the axis scaling"""

    kwargs = dict()
    if which == "x":
        if version_parse(matplotlib.__version__) < version_parse("3.3.0"):
            thresh = "linthreshx"
            base = "basex"
            scale = "linscalex"
        else:
            thresh = "linthresh"
            base = "base"
            scale = "linscale"

        scaler = axes.set_xscale
        limit = axes.set_xlim
    else:
        if version_parse(matplotlib.__version__) < version_parse("3.3.0"):
            thresh = "linthreshy"
            base = "basey"
            scale = "linscaley"
        else:
            thresh = "linthresh"
            base = "base"
            scale = "linscale"

        scaler = axes.set_yscale
        limit = axes.set_ylim

    # Map ticker scales
    if ax_type == "mel":
        mode = "symlog"
        kwargs[thresh] = 1000.0
        kwargs[base] = 2

    elif ax_type in ["cqt", "cqt_hz", "cqt_note", "cqt_svara"]:
        mode = "log"
        kwargs[base] = 2

    elif ax_type in ["log", "fft_note", "fft_svara"]:
        mode = "symlog"
        kwargs[base] = 2
        # kwargs[thresh] = core.note_to_hz(
        #    'C2'
        # )  # in librosa/core.py but I don't think it is needed
        kwargs[scale] = 0.5

    elif ax_type in ["tempo", "fourier_tempo"]:
        mode = "log"
        kwargs[base] = 2
        limit(16, 480)
    else:
        return

    scaler(mode, **kwargs)


def __decorate_axis(axis, ax_type, key="C:maj", Sa=None, mela=None, thaat=None):
    """Configure axis tickers, locators, and labels"""
    if ax_type == "time":
        axis.set_major_formatter(TimeFormatter(unit=None, lag=False))
        axis.set_major_locator(MaxNLocator(prune=None, steps=[1, 1.5, 5, 6, 10]))
        axis.set_label_text("Time")
    elif ax_type in ["mel", "log"]:
        axis.set_major_formatter(ScalarFormatter())
        axis.set_major_locator(SymmetricalLogLocator(axis.get_transform()))
        axis.set_label_text("Hz")


class TimeFormatter(Formatter):
    def __init__(self, lag=False, unit=None):

        if unit not in ["s", "ms", None]:
            raise Exception("Unknown time unit: {}".format(unit))

        self.unit = unit
        self.lag = lag

    def __call__(self, x, pos=None):
        """Return the time format as pos"""

        _, dmax = self.axis.get_data_interval()
        vmin, vmax = self.axis.get_view_interval()

        # In lag-time axes, anything greater than dmax / 2 is negative time
        if self.lag and x >= dmax * 0.5:
            # In lag mode, don't tick past the limits of the data
            if x > dmax:
                return ""
            value = np.abs(x - dmax)
            # Do we need to tweak vmin/vmax here?
            sign = "-"
        else:
            value = x
            sign = ""

        if self.unit == "s":
            s = "{:.3g}".format(value)
        elif self.unit == "ms":
            s = "{:.3g}".format(value * 1000)
        else:
            if vmax - vmin > 3600:
                # Hours viz
                s = "{:d}:{:02d}:{:02d}".format(
                    int(value / 3600.0),
                    int(np.mod(value / 60.0, 60)),
                    int(np.mod(value, 60)),
                )
            elif vmax - vmin > 60:
                # Minutes viz
                s = "{:d}:{:02d}".format(int(value / 60.0), int(np.mod(value, 60)))
            elif vmax - vmin >= 1:
                # Seconds viz
                s = "{:.2g}".format(value)
            else:
                # Milliseconds viz
                s = "{:.3f}".format(value)

        return "{:s}{:s}".format(sign, s)


def __same_axes(x_axis, y_axis, xlim, ylim):
    """Check if two axes are the same, used to determine squared plots"""
    axes_same_and_not_none = (x_axis == y_axis) and (x_axis is not None)
    axes_same_lim = xlim == ylim
    return axes_same_and_not_none and axes_same_lim


# librosa.feature.melspectrogram


def melspectrogram(
    y=None,
    sr=22050,
    S=None,
    n_fft=2048,
    hop_length=512,
    win_length=None,
    window="hann",
    center=True,
    pad_mode="reflect",
    power=2.0,
    **kwargs,
):
    S, n_fft = _spectrogram(
        y=y,
        S=S,
        n_fft=n_fft,
        hop_length=hop_length,
        power=power,
        win_length=win_length,
        window=window,
        center=center,
        pad_mode=pad_mode,
    )

    # Build a Mel filter
    mel_basis = mel(sr, n_fft, **kwargs)

    return np.dot(mel_basis, S)


def _spectrogram(
    y=None,
    S=None,
    n_fft=2048,
    hop_length=512,
    power=1,
    win_length=None,
    window="hann",
    center=True,
    pad_mode="reflect",
):
    if S is not None:
        # Infer n_fft from spectrogram shape
        n_fft = 2 * (S.shape[0] - 1)
    else:
        # Otherwise, compute a magnitude spectrogram from input
        S = (
            np.abs(
                stft(
                    y,
                    n_fft=n_fft,
                    hop_length=hop_length,
                    win_length=win_length,
                    center=center,
                    window=window,
                    pad_mode=pad_mode,
                )
            )
            ** power
        )

    return S, n_fft


def stft(
    y,
    n_fft=2048,
    hop_length=None,
    win_length=None,
    window="hann",
    center=True,
    dtype=None,
    pad_mode="reflect",
):
    # By default, use the entire frame
    if win_length is None:
        win_length = n_fft

    # Set the default hop, if it's not already specified
    if hop_length is None:
        hop_length = int(win_length // 4)

    fft_window = get_window(window, win_length, fftbins=True)

    # Pad the window out to n_fft size
    fft_window = pad_center(fft_window, n_fft)

    # Reshape so that the window can be broadcast
    fft_window = fft_window.reshape((-1, 1))

    # Check audio is valid
    valid_audio(y)

    # Pad the time series so that frames are centered
    if center:
        if n_fft > y.shape[-1]:
            warnings.warn(
                "n_fft={} is too small for input signal of length={}".format(
                    n_fft, y.shape[-1]
                )
            )

        y = np.pad(y, int(n_fft // 2), mode=pad_mode)

    elif n_fft > y.shape[-1]:
        raise Exception(
            "n_fft={} is too large for input signal of length={}".format(
                n_fft, y.shape[-1]
            )
        )

    # Window the time series.
    y_frames = frame(y, frame_length=n_fft, hop_length=hop_length)

    if dtype is None:
        dtype = dtype_r2c(y.dtype)

    # Pre-allocate the STFT matrix
    stft_matrix = np.empty(
        (int(1 + n_fft // 2), y_frames.shape[1]), dtype=dtype, order="F"
    )

    fft = get_fftlib()

    # how many columns can we fit within MAX_MEM_BLOCK?
    n_columns = MAX_MEM_BLOCK // (stft_matrix.shape[0] * stft_matrix.itemsize)
    n_columns = max(n_columns, 1)

    for bl_s in range(0, stft_matrix.shape[1], n_columns):
        bl_t = min(bl_s + n_columns, stft_matrix.shape[1])

        stft_matrix[:, bl_s:bl_t] = fft.rfft(
            fft_window * y_frames[:, bl_s:bl_t], axis=0
        )
    return stft_matrix


def get_window(window, Nx, fftbins=True):
    if callable(window):
        return window(Nx)

    elif isinstance(window, (str, tuple)) or np.isscalar(window):
        # TODO: if we add custom window functions in librosa, call them here

        return scipy.signal.get_window(window, Nx, fftbins=fftbins)

    elif isinstance(window, (np.ndarray, list)):
        if len(window) == Nx:
            return np.asarray(window)

        raise Exception("Window size mismatch: " "{:d} != {:d}".format(len(window), Nx))
    else:
        raise Exception("Invalid window specification: {}".format(window))


def pad_center(data, size, axis=-1, **kwargs):
    kwargs.setdefault("mode", "constant")

    n = data.shape[axis]

    lpad = int((size - n) // 2)

    lengths = [(0, 0)] * data.ndim
    lengths[axis] = (lpad, int(size - n - lpad))

    if lpad < 0:
        raise Exception(
            ("Target size ({:d}) must be " "at least input size ({:d})").format(size, n)
        )

    return np.pad(data, lengths, **kwargs)


def valid_audio(y, mono=True):
    if not isinstance(y, np.ndarray):
        raise Exception("Audio data must be of type numpy.ndarray")

    if not np.issubdtype(y.dtype, np.floating):
        raise Exception("Audio data must be floating-point")

    if mono and y.ndim != 1:
        raise Exception(
            "Invalid shape for monophonic audio: "
            "ndim={:d}, shape={}".format(y.ndim, y.shape)
        )

    elif y.ndim > 2 or y.ndim == 0:
        raise Exception(
            "Audio data must have shape (samples,) or (channels, samples). "
            "Received shape={}".format(y.shape)
        )

    elif y.ndim == 2 and y.shape[0] < 2:
        raise Exception(
            "Mono data must have shape (samples,). " "Received shape={}".format(y.shape)
        )

    if not np.isfinite(y).all():
        raise Exception("Audio buffer is not finite everywhere")

    return True


def frame(x, frame_length, hop_length, axis=-1):
    if not isinstance(x, np.ndarray):
        raise Exception(
            "Input must be of type numpy.ndarray, " "given type(x)={}".format(type(x))
        )

    if x.shape[axis] < frame_length:
        raise Exception(
            "Input is too short (n={:d})"
            " for frame_length={:d}".format(x.shape[axis], frame_length)
        )

    if hop_length < 1:
        raise Exception("Invalid hop_length: {:d}".format(hop_length))

    if axis == -1 and not x.flags["F_CONTIGUOUS"]:
        warnings.warn(
            "librosa.util.frame called with axis={} "
            "on a non-contiguous input. This will result in a copy.".format(axis)
        )
        x = np.asfortranarray(x)
    elif axis == 0 and not x.flags["C_CONTIGUOUS"]:
        warnings.warn(
            "librosa.util.frame called with axis={} "
            "on a non-contiguous input. This will result in a copy.".format(axis)
        )
        x = np.ascontiguousarray(x)

    n_frames = 1 + (x.shape[axis] - frame_length) // hop_length
    strides = np.asarray(x.strides)

    new_stride = np.prod(strides[strides > 0] // x.itemsize) * x.itemsize

    if axis == -1:
        shape = list(x.shape)[:-1] + [frame_length, n_frames]
        strides = list(strides) + [hop_length * new_stride]

    elif axis == 0:
        shape = [n_frames, frame_length] + list(x.shape)[1:]
        strides = [hop_length * new_stride] + list(strides)

    else:
        raise Exception("Frame axis={} must be either 0 or -1".format(axis))

    return as_strided(x, shape=shape, strides=strides)


def dtype_r2c(d, default=np.complex64):
    mapping = {
        np.dtype(np.float32): np.complex64,
        np.dtype(np.float64): np.complex128,
        np.dtype(float): complex,
    }

    # If we're given a complex type already, return it
    dt = np.dtype(d)
    if dt.kind == "c":
        return dt

    # Otherwise, try to map the dtype.
    # If no match is found, return the default.
    return np.dtype(mapping.get(dt, default))


def get_fftlib():
    global __FFTLIB
    return __FFTLIB


def set_fftlib(lib=None):
    global __FFTLIB
    if lib is None:
        from numpy import fft

        lib = fft

    __FFTLIB = lib


set_fftlib(None)


def mel(
    sr,
    n_fft,
    n_mels=128,
    fmin=0.0,
    fmax=None,
    htk=False,
    norm="slaney",
    dtype=np.float32,
):
    if fmax is None:
        fmax = float(sr) / 2

    # Initialize the weights
    n_mels = int(n_mels)
    weights = np.zeros((n_mels, int(1 + n_fft // 2)), dtype=dtype)

    # Center freqs of each FFT bin
    fftfreqs = fft_frequencies(sr=sr, n_fft=n_fft)

    # 'Center freqs' of mel bands - uniformly spaced between limits
    mel_f = mel_frequencies(n_mels + 2, fmin=fmin, fmax=fmax, htk=htk)

    fdiff = np.diff(mel_f)
    ramps = np.subtract.outer(mel_f, fftfreqs)

    for i in range(n_mels):
        # lower and upper slopes for all bins
        lower = -ramps[i] / fdiff[i]
        upper = ramps[i + 2] / fdiff[i + 1]

        # .. then intersect them with each other and zero
        weights[i] = np.maximum(0, np.minimum(lower, upper))

    if norm == "slaney":
        # Slaney-style mel is scaled to be approx constant energy per channel
        enorm = 2.0 / (mel_f[2 : n_mels + 2] - mel_f[:n_mels])
        weights *= enorm[:, np.newaxis]
    else:
        weights = normalize(weights, norm=norm, axis=-1)

    # Only check weights if f_mel[0] is positive
    if not np.all((mel_f[:-2] == 0) | (weights.max(axis=1) > 0)):
        # This means we have an empty channel somewhere
        warnings.warn(
            "Empty filters detected in mel frequency basis. "
            "Some channels will produce empty responses. "
            "Try increasing your sampling rate (and fmax) or "
            "reducing n_mels."
        )

    return weights


def fft_frequencies(sr=22050, n_fft=2048):
    return np.linspace(0, float(sr) / 2, int(1 + n_fft // 2), endpoint=True)


def mel_frequencies(n_mels=128, fmin=0.0, fmax=11025.0, htk=False):
    # 'Center freqs' of mel bands - uniformly spaced between limits
    min_mel = hz_to_mel(fmin, htk=htk)
    max_mel = hz_to_mel(fmax, htk=htk)

    mels = np.linspace(min_mel, max_mel, n_mels)

    return mel_to_hz(mels, htk=htk)


def hz_to_mel(frequencies, htk=False):
    frequencies = np.asanyarray(frequencies)

    if htk:
        return 2595.0 * np.log10(1.0 + frequencies / 700.0)

    # Fill in the linear part
    f_min = 0.0
    f_sp = 200.0 / 3

    mels = (frequencies - f_min) / f_sp

    # Fill in the log-scale part

    min_log_hz = 1000.0  # beginning of log region (Hz)
    min_log_mel = (min_log_hz - f_min) / f_sp  # same (Mels)
    logstep = np.log(6.4) / 27.0  # step size for log region

    if frequencies.ndim:
        # If we have array data, vectorize
        log_t = frequencies >= min_log_hz
        mels[log_t] = min_log_mel + np.log(frequencies[log_t] / min_log_hz) / logstep
    elif frequencies >= min_log_hz:
        # If we have scalar data, heck directly
        mels = min_log_mel + np.log(frequencies / min_log_hz) / logstep

    return mels


def mel_to_hz(mels, htk=False):
    mels = np.asanyarray(mels)

    if htk:
        return 700.0 * (10.0 ** (mels / 2595.0) - 1.0)

    # Fill in the linear scale
    f_min = 0.0
    f_sp = 200.0 / 3
    freqs = f_min + f_sp * mels

    # And now the nonlinear scale
    min_log_hz = 1000.0  # beginning of log region (Hz)
    min_log_mel = (min_log_hz - f_min) / f_sp  # same (Mels)
    logstep = np.log(6.4) / 27.0  # step size for log region

    if mels.ndim:
        # If we have vector data, vectorize
        log_t = mels >= min_log_mel
        freqs[log_t] = min_log_hz * np.exp(logstep * (mels[log_t] - min_log_mel))
    elif mels >= min_log_mel:
        # If we have scalar data, check directly
        freqs = min_log_hz * np.exp(logstep * (mels - min_log_mel))

    return freqs


def normalize(S, norm=np.inf, axis=0, threshold=None, fill=None):
    # Avoid div-by-zero
    if threshold is None:
        threshold = tiny(S)

    elif threshold <= 0:
        raise Exception("threshold={} must be strictly " "positive".format(threshold))

    if fill not in [None, False, True]:
        raise Exception("fill={} must be None or boolean".format(fill))

    if not np.all(np.isfinite(S)):
        raise Exception("Input must be finite")

    # All norms only depend on magnitude, let's do that first
    mag = np.abs(S).astype(np.float)

    # For max/min norms, filling with 1 works
    fill_norm = 1

    if norm == np.inf:
        length = np.max(mag, axis=axis, keepdims=True)

    elif norm == -np.inf:
        length = np.min(mag, axis=axis, keepdims=True)

    elif norm == 0:
        if fill is True:
            raise Exception("Cannot normalize with norm=0 and fill=True")

        length = np.sum(mag > 0, axis=axis, keepdims=True, dtype=mag.dtype)

    elif np.issubdtype(type(norm), np.number) and norm > 0:
        length = np.sum(mag**norm, axis=axis, keepdims=True) ** (1.0 / norm)

        if axis is None:
            fill_norm = mag.size ** (-1.0 / norm)
        else:
            fill_norm = mag.shape[axis] ** (-1.0 / norm)

    elif norm is None:
        return S

    else:
        raise Exception("Unsupported norm: {}".format(repr(norm)))

    # indices where norm is below the threshold
    small_idx = length < threshold

    Snorm = np.empty_like(S)
    if fill is None:
        # Leave small indices un-normalized
        length[small_idx] = 1.0
        Snorm[:] = S / length

    elif fill:
        # If we have a non-zero fill value, we locate those entries by
        # doing a nan-divide.
        # If S was finite, then length is finite (except for small positions)
        length[small_idx] = np.nan
        Snorm[:] = S / length
        Snorm[np.isnan(Snorm)] = fill_norm
    else:
        # Set small values to zero by doing an inf-divide.
        # This is safe (by IEEE-754) as long as S is finite.
        length[small_idx] = np.inf
        Snorm[:] = S / length

    return Snorm


def tiny(x):
    # Make sure we have an array view
    x = np.asarray(x)

    # Only floating types generate a tiny
    if np.issubdtype(x.dtype, np.floating) or np.issubdtype(
        x.dtype, np.complexfloating
    ):
        dtype = x.dtype
    else:
        dtype = np.float32

    return np.finfo(dtype).tiny


# amplitude_to_db


def amplitude_to_db(S, ref=1.0, amin=1e-5, top_db=80.0):
    S = np.asarray(S)

    if np.issubdtype(S.dtype, np.complexfloating):
        warnings.warn(
            "amplitude_to_db was called on complex input so phase "
            "information will be discarded. To suppress this warning, "
            "call amplitude_to_db(np.abs(S)) instead."
        )

    magnitude = np.abs(S)

    if callable(ref):
        # User supplied a function to calculate reference power
        ref_value = ref(magnitude)
    else:
        ref_value = np.abs(ref)

    power = np.square(magnitude, out=magnitude)

    return power_to_db(power, ref=ref_value**2, amin=amin**2, top_db=top_db)


def power_to_db(S, ref=1.0, amin=1e-10, top_db=80.0):
    S = np.asarray(S)

    if amin <= 0:
        raise Exception("amin must be strictly positive")

    if np.issubdtype(S.dtype, np.complexfloating):
        warnings.warn(
            "power_to_db was called on complex input so phase "
            "information will be discarded. To suppress this warning, "
            "call power_to_db(np.abs(D)**2) instead."
        )
        magnitude = np.abs(S)
    else:
        magnitude = S

    if callable(ref):
        # User supplied a function to calculate reference power
        ref_value = ref(magnitude)
    else:
        ref_value = np.abs(ref)

    log_spec = 10.0 * np.log10(np.maximum(amin, magnitude))
    log_spec -= 10.0 * np.log10(np.maximum(amin, ref_value))

    if top_db is not None:
        if top_db < 0:
            raise Exception("top_db must be non-negative")
        log_spec = np.maximum(log_spec, log_spec.max() - top_db)

    return log_spec


# onset_strength


def onset_strength(
    y=None,
    sr=22050,
    S=None,
    lag=1,
    max_size=1,
    ref=None,
    detrend=False,
    center=True,
    feature=None,
    aggregate=None,
    **kwargs,
):
    if aggregate is False:
        raise Exception(
            "aggregate={} cannot be False when computing full-spectrum onset strength."
        )

    odf_all = onset_strength_multi(
        y=y,
        sr=sr,
        S=S,
        lag=lag,
        max_size=max_size,
        ref=ref,
        detrend=detrend,
        center=center,
        feature=feature,
        aggregate=aggregate,
        channels=None,
        **kwargs,
    )

    return odf_all[0]


def onset_strength_multi(
    y=None,
    sr=22050,
    S=None,
    n_fft=2048,
    hop_length=512,
    lag=1,
    max_size=1,
    ref=None,
    detrend=False,
    center=True,
    feature=None,
    aggregate=None,
    channels=None,
    **kwargs,
):
    if feature is None:
        feature = melspectrogram
        kwargs.setdefault("fmax", 11025.0)

    if aggregate is None:
        aggregate = np.mean

    if lag < 1 or not isinstance(lag, (int, np.integer)):
        raise Exception("lag must be a positive integer")

    if max_size < 1 or not isinstance(max_size, (int, np.integer)):
        raise Exception("max_size must be a positive integer")

    # First, compute mel spectrogram
    if S is None:
        S = np.abs(feature(y=y, sr=sr, n_fft=n_fft, hop_length=hop_length, **kwargs))

        # Convert to dBs
        S = power_to_db(S)

    # Ensure that S is at least 2-d
    S = np.atleast_2d(S)

    # Compute the reference spectrogram.
    # Efficiency hack: skip filtering step and pass by reference
    # if max_size will produce a no-op.
    if ref is None:
        if max_size == 1:
            ref = S
        else:
            ref = scipy.ndimage.maximum_filter1d(S, max_size, axis=0)
    elif ref.shape != S.shape:
        raise Exception(
            "Reference spectrum shape {} must match input spectrum {}".format(
                ref.shape, S.shape
            )
        )

    # Compute difference to the reference, spaced by lag
    onset_env = S[:, lag:] - ref[:, :-lag]

    # Discard negatives (decreasing amplitude)
    onset_env = np.maximum(0.0, onset_env)

    # Aggregate within channels
    pad = True
    if channels is None:
        channels = [slice(None)]
    else:
        pad = False

    if aggregate:
        onset_env = sync(onset_env, channels, aggregate=aggregate, pad=pad, axis=0)

    # compensate for lag
    pad_width = lag
    if center:
        # Counter-act framing effects. Shift the onsets by n_fft / hop_length
        pad_width += n_fft // (2 * hop_length)

    onset_env = np.pad(onset_env, ([0, 0], [int(pad_width), 0]), mode="constant")

    # remove the DC component
    if detrend:
        onset_env = scipy.signal.lfilter([1.0, -1.0], [1.0, -0.99], onset_env, axis=-1)

    # Trim to match the input duration
    if center:
        onset_env = onset_env[:, : S.shape[1]]

    return onset_env


def sync(data, idx, aggregate=None, pad=True, axis=-1):
    if aggregate is None:
        aggregate = np.mean

    shape = list(data.shape)

    if np.all([isinstance(_, slice) for _ in idx]):
        slices = idx
    elif np.all([np.issubdtype(type(_), np.integer) for _ in idx]):
        slices = index_to_slice(np.asarray(idx), 0, shape[axis], pad=pad)
    else:
        raise Exception("Invalid index set: {}".format(idx))

    agg_shape = list(shape)
    agg_shape[axis] = len(slices)

    data_agg = np.empty(
        agg_shape, order="F" if np.isfortran(data) else "C", dtype=data.dtype
    )

    idx_in = [slice(None)] * data.ndim
    idx_agg = [slice(None)] * data_agg.ndim

    for i, segment in enumerate(slices):
        idx_in[axis] = segment
        idx_agg[axis] = i
        data_agg[tuple(idx_agg)] = aggregate(data[tuple(idx_in)], axis=axis)

    return data_agg


def index_to_slice(idx, idx_min=None, idx_max=None, step=None, pad=True):
    # First, normalize the index set
    idx_fixed = fix_frames(idx, idx_min, idx_max, pad=pad)

    # Now convert the indices to slices
    return [slice(start, end, step) for (start, end) in zip(idx_fixed, idx_fixed[1:])]


def fix_frames(frames, x_min=0, x_max=None, pad=True):
    frames = np.asarray(frames)

    if np.any(frames < 0):
        raise Exception("Negative frame index detected")

    if pad and (x_min is not None or x_max is not None):
        frames = np.clip(frames, x_min, x_max)

    if pad:
        pad_data = []
        if x_min is not None:
            pad_data.append(x_min)
        if x_max is not None:
            pad_data.append(x_max)
        frames = np.concatenate((pad_data, frames))

    if x_min is not None:
        frames = frames[frames >= x_min]

    if x_max is not None:
        frames = frames[frames <= x_max]

    return np.unique(frames).astype(int)


def spectral_contrast(
    y=None,
    sr=22050,
    S=None,
    n_fft=2048,
    hop_length=512,
    win_length=None,
    window="hann",
    center=True,
    pad_mode="reflect",
    freq=None,
    fmin=200.0,
    n_bands=6,
    quantile=0.02,
    linear=False,
):
    S, n_fft = _spectrogram(
        y=y,
        S=S,
        n_fft=n_fft,
        hop_length=hop_length,
        win_length=win_length,
        window=window,
        center=center,
        pad_mode=pad_mode,
    )

    # Compute the center frequencies of each bin
    if freq is None:
        freq = fft_frequencies(sr=sr, n_fft=n_fft)

    freq = np.atleast_1d(freq)

    if freq.ndim != 1 or len(freq) != S.shape[0]:
        raise Exception("freq.shape mismatch: expected " "({:d},)".format(S.shape[0]))

    if n_bands < 1 or not isinstance(n_bands, int):
        raise Exception("n_bands must be a positive integer")

    if not 0.0 < quantile < 1.0:
        raise Exception("quantile must lie in the range (0, 1)")

    if fmin <= 0:
        raise Exception("fmin must be a positive number")

    octa = np.zeros(n_bands + 2)
    octa[1:] = fmin * (2.0 ** np.arange(0, n_bands + 1))

    if np.any(octa[:-1] >= 0.5 * sr):
        raise Exception(
            "Frequency band exceeds Nyquist. " "Reduce either fmin or n_bands."
        )

    valley = np.zeros((n_bands + 1, S.shape[1]))
    peak = np.zeros_like(valley)

    for k, (f_low, f_high) in enumerate(zip(octa[:-1], octa[1:])):
        current_band = np.logical_and(freq >= f_low, freq <= f_high)

        idx = np.flatnonzero(current_band)

        if k > 0:
            current_band[idx[0] - 1] = True

        if k == n_bands:
            current_band[idx[-1] + 1 :] = True

        sub_band = S[current_band]

        if k < n_bands:
            sub_band = sub_band[:-1]

        # Always take at least one bin from each side
        idx = np.rint(quantile * np.sum(current_band))
        idx = int(np.maximum(idx, 1))

        sortedr = np.sort(sub_band, axis=0)

        valley[k] = np.mean(sortedr[:idx], axis=0)
        peak[k] = np.mean(sortedr[-idx:], axis=0)

    if linear:
        return peak - valley
    else:
        return power_to_db(peak) - power_to_db(valley)


# db_to_amplitude


def db_to_amplitude(S_db, ref=1.0):
    return db_to_power(S_db, ref=ref**2) ** 0.5


def db_to_power(S_db, ref=1.0):
    return ref * np.power(10.0, 0.1 * S_db)


# spectral_centroid


def spectral_centroid(
    y=None,
    sr=22050,
    S=None,
    n_fft=2048,
    hop_length=512,
    freq=None,
    win_length=None,
    window="hann",
    center=True,
    pad_mode="reflect",
):
    S, n_fft = _spectrogram(
        y=y,
        S=S,
        n_fft=n_fft,
        hop_length=hop_length,
        win_length=win_length,
        window=window,
        center=center,
        pad_mode=pad_mode,
    )

    if not np.isrealobj(S):
        raise Exception("Spectral centroid is only defined " "with real-valued input")
    elif np.any(S < 0):
        raise Exception(
            "Spectral centroid is only defined " "with non-negative energies"
        )

    # Compute the center frequencies of each bin
    if freq is None:
        freq = fft_frequencies(sr=sr, n_fft=n_fft)

    if freq.ndim == 1:
        freq = freq.reshape((-1, 1))

    # Column-normalize S
    return np.sum(freq * normalize(S, norm=1, axis=0), axis=0, keepdims=True)



"""Make annotated sounds"""

import random
import itertools
from itertools import starmap
from functools import partial

import pandas as pd
import numpy as np
from sklearn.preprocessing import MinMaxScaler

from i2 import Pipe
from slink import dict_generator, GetFromIter, Repeater

from loopyng.gen.sine_mix import dflt_wf_params_to_wf


def session_phase_rpm_temparature(
    n_sessions=2,
    n_phases_per_session=3,
    n_blocks_per_phase=2,
    average_block_duration=21 * 2048,
    wf_params_cols=(
        "session",
        "phase",
        "rpm",
        "temperature",
    ),  # TODO: perhaps an exclusion list is more general
    annots_df_to_wf_params=MinMaxScaler().fit_transform,
    params_and_duration_to_wf=lambda p, duration: dflt_wf_params_to_wf(
        p, n_samples=duration
    ),
):
    wf_params_cols = list(wf_params_cols)
    f = dict_generator(
        # make n_sessions copies of the dict so far... i.e. empty dict
        Repeater(n_sessions),
        # --> {}, {}
        #
        # for each, call function and assign to session key
        dict(session=GetFromIter(itertools.count())),
        # --> {'session': 0}, {'session': 1}
        #
        # make n_phases_per_session copies of each
        Repeater(n_phases_per_session),
        # --> {'session': 0}, {'session': 0}, {'session': 0}, ...
        # ... {'session': 1}, {'session': 1}, {'session': 1}
        #
        # for each, make a phase using given (indep) function
        dict(phase=GetFromIter(itertools.cycle(range(n_phases_per_session)))),
        #
        # etc.
        Repeater(n_blocks_per_phase),
        dict(
            rpm=Pipe(partial(random.uniform, 200, 800), int),  # make a randome rpm
            temperature=lambda rpm: int(
                rpm * random.uniform(0.5, 2) / 50
            ),  # use the rpm to compute temperature
            duration=lambda: int(average_block_duration * random.random()),
        ),
    )
    annots_df = pd.DataFrame(f())
    annots_df["timestamp"] = list(itertools.accumulate(annots_df["duration"]))
    wf_params = annots_df_to_wf_params(annots_df[wf_params_cols])

    wf = np.hstack(
        list(starmap(params_and_duration_to_wf, zip(wf_params, annots_df["duration"])))
    )

    return annots_df.to_dict(orient="records"), wf



"""
Plot utilities for visualizing audio waveforms and spectrograms.

Provides functions for displaying waveforms with proper time axes and playback options.
"""

from inspect import getmodule
import matplotlib.pylab as plt
from numpy import linspace
from loopyng.utils.date_ticks import str_ticks

DFLT_FIGSIZE_FOR_WF_PLOTS = (22, 5)
DFLT_SR = 44100


def getmodulename(obj, default=""):
    """Get name of module of object"""
    return getattr(getmodule(obj), "__name__", default)


# def plot_wf(wf, sr=None, figsize=(20, 6), **kwargs):
#     if figsize is not None:
#         plt.figure(figsize=figsize)
#     if sr is not None:
#         plt.plot(linspace(start=0, stop=len(wf) / float(sr), num=len(wf)), wf, **kwargs)
#     else:
#         plt.plot(wf, **kwargs)


def plot_wf(
    wf, sr=None, figsize=DFLT_FIGSIZE_FOR_WF_PLOTS, offset_s=0, ax=None, **kwargs
):
    if figsize is not None:
        plt.figure(figsize=figsize)
    _ax = ax or plt
    if sr is not None:
        _ax.plot(
            offset_s + linspace(start=0, stop=len(wf) / float(sr), num=len(wf)),
            wf,
            **kwargs
        )
        plt.margins(x=0)
    else:
        _ax.plot(wf, **kwargs)
        plt.margins(x=0)
        return
    if _ax == plt:
        _xticks, _ = plt.xticks()
        plt.xticks(_xticks, str_ticks(ticks=_xticks, ticks_unit=1))
        plt.margins(x=0)
    else:
        _xticks = _ax.get_xticks()
        _ax.set_xticks(_xticks)
        _ax.set_xticklabels(str_ticks(ticks=_xticks, ticks_unit=1))
        plt.margins(x=0)


def disp_wf(wf, sr=DFLT_SR, autoplay=False, wf_plot_func=plt.specgram):
    if wf_plot_func is not None:
        if getmodulename(wf_plot_func, "").startswith("matplotlib"):
            plt.figure(figsize=DFLT_FIGSIZE_FOR_WF_PLOTS)
        wf_plot_func(wf, sr)
    try:
        from IPython.display import Audio

        return Audio(data=wf, rate=sr, autoplay=autoplay)
    except:
        pass



"""
A few general utils
"""

from warnings import warn
from inspect import getmodule
from typing import Iterable


def simple_chunker(a: Iterable, chk_size: int):
    """Generate fixed sized non-overlapping chunks of an iterable ``a``.

    >>> list(simple_chunker(range(7), 3))
    [(0, 1, 2), (3, 4, 5)]
    """
    return zip(*([iter(a)] * chk_size))


def getmodulename(obj, default=""):
    """Get name of module of object"""
    return getattr(getmodule(obj), "__name__", default)


class ModuleNotFoundErrorNiceMessage:
    def __init__(self, msg=None):
        self.msg = msg

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is ModuleNotFoundError:
            msg = (
                self.msg
                or f"""
It seems you don't have required `{exc_val.name}` package for this Store.
Try installing it by running:

    pip install {exc_val.name}

in your terminal.
For more information: https://pypi.org/project/{exc_val.name}
            """
            )
            warn(msg)


class ModuleNotFoundIgnore:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is ModuleNotFoundError:
            pass
        return True
